
import inspect
import threading
from collections import deque
from weakref import ref as wkref, WeakSet, WeakValueDictionary
import functools


class AcceptError(PermissionError): pass


def ref(obj, callback=None):
    if hasattr(obj, "get_weakref"):
        return obj.get_weakref(callback)
    return wkref(obj, callback)


class PrefilledFunction:
    def __init__(self, function, *args, **kwargs):
        assert callable(function), "function must be callable"

        self._function = function
        self._args = args
        self._kwargs = kwargs

    def __call__(self, *args, **kwargs):

        return self._function(*self._args, *args, **self._kwargs, **kwargs)


class PackedFunctions:
    """
    This object allow tou to execute multiple function from one call

    WARNING : if you enter parameters, it will be distributed to all of the packed functions
    This means you cannot pack two functions if they each require differents parameters
    You cannot remove a function from a PackedFunctions
    """
    def __init__(self, *functions):

        self._functions = list()
        for func in functions:
            self.add(func)

    def __call__(self, *args, **kwargs):

        for func in self._functions:
            func(*args, **kwargs)

    def __str__(self):

        return "PackedFunctions{}".format(tuple(self._functions))

    def add(self, func, owner=None):
        """
        Owner parameter is usefull for remove
        It allow you to remove an element who own a method in a PackedFunctions
        because if you don't, the owner will not be deleted
        """

        assert callable(func)
        if owner is not None:
            setattr(func, "_owner", owner)
            owner.listof_packedfunctions.add(self)
        self._functions.append(func)

    def clear(self):

        self._functions.clear()

    def remove(self, func):

        self._functions.remove(func)

    def remove_funcs_of(self, owner):

        # funcs_to_remove = []
        for func in tuple(self._functions):
            if hasattr(func, "_owner"):
                if func._owner == owner:
                    # funcs_to_remove.append(func)
                    self._functions.remove(func)
        # for func in funcs_to_remove:


class Object:
    def __init__(self, dict=None, **kwargs):
        if dict is None:
            dict = kwargs.items()
        for key, value in dict:
            self.__setattr__(key, value)

    def __str__(self) -> str:
        return "<{}({})>".format(self.__class__.__name__, str(self.__dict__)[1:-1])


class EditableTuple:
    """
    An EditableTuple has a fixed length but its elements can be changed

    WARNING : an EditableTuple is not a tuple instance
    """

    def __init__(self, tup):

        self._tup = tuple(tup)
    def __contains__(self, key):
        """ Return key in self. """
        return self._tup.__contains__(key)

    def __eq__(self, value):
        """ Return self==value. """
        return self._tup.__eq__(value)

    def __ge__(self, value):
        """ Return self>=value. """
        return self._tup.__ge__(value)

    def __getitem__(self, key):
        """ Return self[key]. """
        return self._tup.__getitem__(key)

    def __gt__(self, value):
        """ Return self>value. """
        return self._tup.__gt__(value)

    def __hash__(self):
        """ Return hash(self). """
        return self._tup.__hash__()

    def __iter__(self):
        """ Implement iter(self). """
        return self._tup.__iter__()

    def __len__(self):
        """ Return len(self). """
        return self._tup.__len__()

    def __le__(self, value):
        """ Return self<=value. """
        return self._tup.__le__(value)

    def __lt__(self, value):
        """ Return self<value. """
        return self._tup.__lt__(value)

    def __ne__(self, value):
        """ Return self!=value. """
        return self._tup.__ne__(value)

    def __repr__(self):
        """ Return repr(self). """
        return self._tup.__repr__()

    def __setitem__(self, key, value):
        """ Set self[key] to value. """

        change = lambda k, v: value if k == key else v
        # tup = tuple([change(k, v) for k, v in enumerate(self._tup)])
        # if tup == self._tup:
        #     raise IndexError("tuple index out of range (given index : %s)" % key)
        self._tup = tuple([change(k, v) for k, v in enumerate(self._tup)])

    def __str__(self):
        """ Return str(self). """
        return self._tup.__str__()

    def count(self, value):
        """ T.count(value) -> integer -- return number of occurrences of value """
        return self._tup.count(value)

    def index(self, value, start=None, stop=None):
        """
        T.index(value, [start, [stop]]) -> integer -- return first index of value.
        Raises ValueError if the value is not present.
        """
        return self._tup.index(value, start, stop)


class TypedDeque(deque):
    """
    A TypedDeque is a deque who can only contain items of type ItemsClass
    deque : a list-like sequence optimized for data accesses near its endpoints
    """
    def __init__(self, ItemsClass, seq=(), maxlen=None):
        """
        :param ItemsClass: the type for all deque items
        :param seq: an items sequence
        :param maxlen: the max deque length
        """

        assert inspect.isclass(ItemsClass), "ItemsClass must be a class"
        assert hasattr(seq, "__iter__"), "seq must be an iterable"
        assert maxlen is None or isinstance(maxlen, int), "maxlen must be an integer"

        self.ItemsClass = ItemsClass
        self.msg_item_type_error = "Only {} objects are accepted in this list".format(self.ItemsClass.__name__)

        for item in seq:
            self._check(item)
        deque.__init__(self, seq, maxlen)

    def __setitem__(self, index, p_object):
        self._check(p_object)

        deque.__setitem__(self, index, p_object)

    def __repr__(self):
        return "<TypedDeque({}):{}>".format(self.ItemsClass.__name__,
                                           "[{}]".format(", ".join(list((item.__str__() for item in self)))))

    def __str__(self):
        return "[{}]".format(", ".join(list((item.__str__() for item in self)))) + (", maxlen={}".format(self.maxlen) if self.maxlen else "")

    def _check(self, p_object):
        if not self.accept(p_object):
            raise AcceptError(self.msg_item_type_error)

    def accept(self, p_object):
        return isinstance(p_object, self.ItemsClass)

    def append(self, p_object):
        self._check(p_object)

        deque.append(self, p_object)

    def appendleft(self, p_object):
        self._check(p_object)

        deque.appendleft(self, p_object)

    def extend(self, iterable):
        for p_object in iterable:
            self.append(p_object)

    def extendleft(self, iterable):
        for p_object in iterable:
            self.appendleft(p_object)

    def insert(self, index, p_object):
        self._check(p_object)

        deque.insert(self, index, p_object)


class TypedDict(dict):
    def __init__(self, KeysClass, ValuesClass, seq={}, **kwargs):
        """
        Create a dict who can only contain keys of type keys_class
        and values of type values_class

        :param KeysClass: the type for all dict keys
        :type KeysClass: class
        :param ValuesClass: the type for all dict values
        :type ValuesClass: class
        :param seq: an items sequence
        :param kwargs: an items dictionnary
        """
        assert inspect.isclass(KeysClass), "KeysClass must be a class"
        assert inspect.isclass(ValuesClass), "ValuesClass must be a class"
        assert isinstance(seq, dict), "Optionnal seq must be a dictionnary"

        self.KeysClass = KeysClass
        self.ValuesClass = ValuesClass
        self.msg_key_type_error = "Only {} objects are accepted as key in this dict" \
                                  "".format(self.KeysClass.__name__)
        self.msg_value_type_error = "Only {} objects are accepted as value in this dict" \
                                    "".format(self.ValuesClass.__name__)

        for key, value in seq.items():
            self._checkkey(key)
            self._checkvalue(value)
        dict.__init__(self, seq, **kwargs)

    def __setitem__(self, *args, **kwargs):
        try:
            self._checkkey(args[0])
            self._checkvalue(args[1])
            dict.__setitem__(self, *args, **kwargs)
        except Exception as e:
            raise Exception("This method is not properly coded, " + str(e))

    def __repr__(self):
        return "<TypedDict({}, {}):{}>" \
               "".format(self.KeysClass.__name__, self.ValuesClass.__name__, dict.__str__(self))

    def __str__(self):
        return "{"+", ".join("{}:{}".format(i, o) for i, o in self.__dict__.items())+"}"

    def _checkkey(self, key):
        if not self.acceptkey(key):
            raise AcceptError(self.msg_key_type_error)

    def _checkvalue(self, value):
        if not self.acceptvalue(value):
            raise AcceptError(self.msg_value_type_error)

    def acceptkey(self, item):
        return isinstance(item, self.KeysClass)

    def acceptvalue(self, item):
        return isinstance(item, self.ValuesClass)

    def setdefault(self, k, d=None):
        """
        If k not in D: set D[k]=d
        If d is None:  set d=D.ValuesClass()
        Return D[k]

        WARNING: ValuesClass might need arguments

        :param k: a key
        :param d: the key value if key isn't in the dict yet
        :return: D[k]
        """
        if d is None:
            d = self.ValuesClass()

        self._checkkey(k)
        self._checkvalue(v)

        return dict.setdefault(self, k, d)

    def update(self, E=None, **F):
        """
        Set D[k] = v for k, v in E.items() or F.items()

        D.update(key1=1, key2=2) <=> D.update({'key1':1, 'key2':2})

        :param E: a dictionnary
        :param F: optionnal - a dictionnary (made by keywords)
        :return: None
        """
        if E is None:
            E = F
        assert isinstance(E, dict), "E must be a dictionnary"

        for k, v in E.items():
            self[k] = v


class TypedList(list):

    def __init__(self, *ItemsClass, seq=()):
        """
        Create a list who can only contain items of type ItemsClass
        :param ItemsClass: the type for all list items
        :type ItemsClass: class
        :param seq: an items sequence
        """
        list.__init__(self, seq)
        self.set_ItemsClass(*ItemsClass)

    def __setitem__(self, index, p_object):
        self._check(p_object)

        list.__setitem__(self, index, p_object)

    def __repr__(self):
        return "<{}(ItemsClass:{}, {}>".format(self.__class__.__name__, self.ItemsClass_name,
                                    "[{}]".format(", ".join(list((item.__str__() for item in self)))))

    def __str__(self):
        return "[{}]".format(", ".join(list((item.__str__() for item in self))))

    def _check(self, item):
        if not self.accept(item):
            raise AcceptError(self.msg_item_type_error.format(item.__class__.__name__))

    def accept(self, item):
        # print("accept", item.__class__.__name__, self.ItemsClass_name, isinstance(item, self.ItemsClass))
        if inspect.isclass(item):
            return issubclass(item, self.ItemsClass)
        return isinstance(item, self.ItemsClass)

    def append(self, p_object):
        self._check(p_object)

        list.append(self, p_object)

    def extend(self, iterable):
        for p_object in iterable:
            self.append(p_object)

    def insert(self, index, p_object):
        self._check(p_object)

        list.insert(self, index, p_object)

    def set_ItemsClass(self, *ItemsClass):

        # try:
        for Class in ItemsClass:
            assert inspect.isclass(Class), "ItemsClass must be a class or a list of class"
        name = "(%s)" % ", ".join(Class.__name__ for Class in ItemsClass)
        # except TypeError:
        #     assert inspect.isclass(ItemsClass), "ItemsClass must be a class or a list of class"
        #     name = ItemsClass.__name__
        self.ItemsClass = ItemsClass
        self.ItemsClass_name = name
        self.msg_item_type_error = "Only {} objects are accepted in this list".format(self.ItemsClass_name) + \
            " (wrong object class:{})"
        for item in self:
            self._check(p_object)


class SortedTypedList(TypedList):
    """
    Create an list of Focusable components ordered by there positions
    """

    def __init__(self, *ItemsClass, sort_key, seq=()):

        TypedList.__init__(self, *ItemsClass, seq)
        self._sort_key = sort_key

    def accept(self, item):
        return hasattr(item, self._sort_key) and super().accept(item)

    def append(self, item):
        assert item not in self

        item_key = getattr(item, self._sort_key)
        for i, item2 in enumerate(self):
            item2_key = getattr(item2, self._sort_key)
            if item_key < item2_key:
                super().insert(i, item)
                break
        if not item in self:
            super().append(item)

        assert list(self) == sorted(self, key=lambda o: getattr(o, self._sort_key))

    def insert(self, index, comp):
        raise PermissionError("Cannot insert item on a sorted list")

    def sort(self):
        super().sort(key=lambda o: getattr(o, self._sort_key))


class TypedSet(set):
    """
    A TypedSet is a unordered collection of unique elements
    who can only contain items of type ItemsClass

    seq is the optionnal initial sequence
    """
    def __init__(self, ItemsClass, seq=()):

        assert inspect.isclass(ItemsClass), "ItemsClass must be a class"
        assert hasattr(seq, "__iter__"), "seq must be an iterable"
        self.ItemsClass = ItemsClass
        self.msg_item_type_error = "Only {} objects are accepted in this list".format(self.ItemsClass.__name__)
        for item in seq:
            self._check(item)
        set.__init__(self, seq)

    def __repr__(self):
        return super().__repr__()
        # return "<TypedSet({}):{}>".format(self.ItemsClass.__name__, set.__str__(self))

    def __str__(self):
        return super().__str__()
        # return "{{}}".format(", ".join(list((item.__str__() for item in self))))

    def _check(self, item):
        if not self.accept(item):
            raise AcceptError(self.msg_item_type_error)

    def accept(self, item):
        return isinstance(item, self.ItemsClass)

    def add(self, item):
        """
        Add an item to a set.
        This has no effect if the item is already present.

        :param item: an item of type self.ItemsClass
        :return: None
        """
        self._check(item)

        set.add(self, item)

    def update(self, *args):
        """
        Add all items from args into a set
        if the items already is in the set, it isn't added

        S.update({1, 2, 3})
        S.update({1}, {2}, {3})

        :param args: a sequence of items sequence
        :return: None
        """
        for seq in args:
            for item in seq:
                self.add(item)


class History(TypedDeque):
    """
    An History is a TypedDeque whith a fixed size
    You can only :
        - append a new element in the history
        - read the history
    When you append a new element to the history, if it is full,
    the oldest element is removed
    The oldest element is positionned to the left

    Exemple :

            print(history) -> [2, 3, 4, 5, 6, 7, 8, 9]
                               ^                    ^
                        oldest element       newest element
    """

    def __init__(self, ItemsClass, maxlen, seq=()):
        TypedDeque.__init__(self, ItemsClass, seq, maxlen)

    def __delitem__(self, *args, **kwargs):
        raise PermissionError("Cannot use __delitem__ on an History")

    def __iadd__(self, *args, **kwargs):
        raise PermissionError("Cannot use __iadd__ on an History")

    def __imul__(self, *args, **kwargs):
        raise PermissionError("Cannot use __imul__ on an History")

    def __setitem__(self, index, p_object):
        raise PermissionError("Cannot use __setitem__ on an History")

    def appendleft(self, p_object):
        raise PermissionError("Cannot use appendleft on an History")

    def extendleft(self, iterable):
        raise PermissionError("Cannot use extendleft on an History")

    def insert(self, index, p_object):
        raise PermissionError("Cannot use insert on an History")

    def pop(self, *args, **kwargs):
        raise PermissionError("Cannot use pop on an History")

    def popleft(self, *args, **kwargs):
        raise PermissionError("Cannot use popleft on an History")

    def reverse(self):
        raise PermissionError("Cannot use reverse on an History")

    def rotate(self, *args, **kwargs):
        raise PermissionError("Cannot use rotate on an History")


class WeakList(list):
    """
    Create a TypedList who only store weak references to objects
    code from : https://stackoverflow.com/questions/677978/weakref-list-in-python
    """
    def __init__(self, seq=()):
        list.__init__(self)
        self._refs = []
        self._dirty=False
        for x in seq: self.append(x)

    def _mark_dirty(self, wref):
        self._dirty = True

    def flush(self):
        self._refs = [x for x in self._refs if x() is not None]
        self._dirty=False

    def __eq__(self, other):

        return self is other

    def __getitem__(self, idx):
        if self._dirty: self.flush()
        if type(idx) == slice:
            return list(ref() for ref in self._refs[idx])
        return self._refs[idx]()

    def __iter__(self):
        if self._dirty: self.flush()
        for ref in self._refs:
            yield ref()

    def __repr__(self):
        return "WeakList(%r)" % list(self)

    def __str__(self):
        return "[{}]".format(", ".join(list((item.__str__() for item in list(self)))))

    def __len__(self):
        if self._dirty: self.flush()
        return len(self._refs)

    def __setitem__(self, idx, obj):
        if isinstance(idx, slice):
            self._refs[idx] = [ref(obj, self._mark_dirty) for x in obj]
        else:
            self._refs[idx] = ref(obj, self._mark_dirty)

    def __delitem__(self, idx):
        del self._refs[idx]

    def append(self, obj):
        self._refs.append(ref(obj, self._mark_dirty))

    def count(self, obj):
        return list(self).count(obj)

    def extend(self, items):
        for x in items: self.append(x)

    def index(self, obj):
        return list(self).index(obj)

    def insert(self, idx, obj):
        self._refs.insert(idx, ref(obj, self._mark_dirty))

    def pop(self, idx):
        if self._dirty: self.flush()
        obj=self._refs[idx]()
        del self._refs[idx]
        self.flush()
        return obj

    def remove(self, obj):
        if self._dirty: self.flush() # Ensure all valid.
        for i, x in enumerate(self):
            if x == obj:
                del self[i]
        self.flush()

    def reverse(self):
        self._refs.reverse()

    def sort(self, key=None, reverse=False):
        if self._dirty: self.flush()
        if key is not None:
            key = lambda x, key=key: key(x())
        else:
            key = lambda x: x()
        self._refs.sort(key=key, reverse=reverse)

    def __add__(self, other):
        l = WeakList(self)
        l.extend(other)
        return l

    def __iadd__(self, other):
        self.extend(other)
        return self

    def __contains__(self, obj):
        return obj in list(self)

    def __mul__(self, n):
        return WeakList(list(self)*n)

    def __imul__(self, n):
        self._refs *= n
        return self

    def __reversed__(self):
        for ref in self._refs.__reversed__():
            yield ref()


class WeakTypedList(TypedList, WeakList):

    def __init__(self, *ItemsClass, seq=()):

        WeakList.__init__(self, seq=seq)
        TypedList.__init__(self, *ItemsClass)

    def __eq__(self, other):

        return self is other

    def __setitem__(self, index, p_object):
        self._check(p_object)

        WeakList.__setitem__(self, index, p_object)

    def __str__(self):
        return WeakList.__str__(self)

    def append(self, p_object):
        self._check(p_object)

        WeakList.append(self, p_object)

    def extend(self, iterable):
        for p_object in iterable:
            self.append(p_object)

    def insert(self, index, p_object):
        self._check(p_object)

        WeakList.insert(self, index, p_object)


class WeakTypedSet(WeakSet, TypedSet):
    """
    A TypedSet is a unordered collection of unique elements
    who can only contain items of type ItemsClass

    seq is the optionnal initial sequence

    WARNING : elements of a WeakTypedSet object are stored in WeakTypedSet.data
    """
    def __init__(self, ItemsClass, seq=()):

        WeakSet.__init__(self, data=seq)
        TypedSet.__init__(self, ItemsClass=ItemsClass)

    def __eq__(self, other):

        return self is other

    def __repr__(self):
        return super().__repr__()
        # return "<WeakTypedSet({}):{}>".format(self.ItemsClass.__name__, set.__str__(self))

    def __str__(self):
        return super().__str__()
        # return "{{}}".format(", ".join(list((item.__str__() for item in self))))

    def add(self, item):
        """
        Add an item to a set.
        This has no effect if the item is already present.

        :param item: an item of type self.ItemsClass
        :return: None
        """
        self._check(item)
        WeakSet.add(self, item)

    def update(self, *args):
        """
        Add all items from args into a set
        if the items already is in the set, it isn't added

        S.update({1, 2, 3})
        S.update({1}, {2}, {3})

        :param args: a sequence of items sequence
        :return: None
        """
        for seq in args:
            for item in seq:
                self.add(item)


class WeakSortedTypedList_TBR(WeakTypedList):
    """
    Create an list of Focusable components ordered by a key
    WARNING : the list will be sorted after an append(elt) and after a sort(),
              but it is not updating in real-time
    """
    # TODO : remove this class, because it creates logic issues (we think the list is always sorted but it's not)

    def __init__(self, *ItemsClass, key=None, seq=()):

        WeakTypedList.__init__(self, *ItemsClass, seq)
        self._key = key
        if key is not None:
            def cannot_insert(index, p_object):
                raise PermissionError("Cannot insert item on a sorted list")
            self.insert = cannot_insert

    def append(self, item):
        assert item not in self

        super().append(item)
        if self._key is not None:
            self.sort()

    def sort(self):
        if self._dirty: self.flush()
        if self._key is not None:
            super().sort(key=self._key)
        else:
            super().sort()


""" --- Unit tests for WeakList and WeakTypedList ---

class Obj:
    def __init__(self, num):
        self.id = num
    def __str__(self):
        return "Obj(%s)" % self.id


weak_list = WeakList()
weak_list2 = WeakList()
strong_list = []
weak_typed_list = WeakTypedList(ItemsClass=Obj)

weak_typed_list.append(Obj(4))
assert len(weak_typed_list) == 0

for i in range(4):
    obj = Obj(i)
    strong_list.append(obj)
    weak_list.append(obj)
    weak_list2.append(obj)
    weak_typed_list.append(obj)

print(weak_list)
print(len(weak_list))
assert len(weak_list) == 4

del strong_list[2]
assert len(weak_list) == 3
for obj in weak_list:
    print(obj)
print(weak_list, weak_list2, weak_typed_list)
print(strong_list)
"""

"""
# --- Unit tests for WeakSet and WeakTypedSet ---

class Obj:
    def __init__(self, num):
        self.id = num
    def __str__(self):
        return "Obj(%s)" % self.id
    def __del__(self):
        print("DELETE : " + str(self))


obj = Obj(3)
weak_set = WeakSet()
weak_typed_set = WeakTypedSet(Obj)

weak_set.add(obj)
weak_typed_set.add(obj)

print("weak_set :", weak_set)
print("weak_typed_set :", weak_typed_set)
del obj
print("weak_set", weak_set)
print("weak_typed_set :", weak_typed_set)
"""

def get_name(obj):
    for k, v in globals().items():
        if k == "obj":
            continue
        if v == obj:
            return k
    return None

    """ --- get_name testing ---
    abc = 3
    print(get_name(abc))
    """


def debug(func):
    functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            raise e
    return wrapper






from baopig import *


class UT_TextLabel(Zone):
    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)

        label = TextLabel(self, "Hello\nHow are you dear ?", background_color="yellow", pos=(10, 10),
                          width=100, height=200, align_mode="center")
        Button(self, "Click", pos=(120, 10), height=100)


ut_scenes = [
    UT_TextLabel,
]


if __name__ == "__main__":
    from baopig.unit_tests.TesterScene import TesterScene
    app = Application()
    for scene in ut_scenes:
        TesterScene(app, scene)
    app.launch()





from baopig.pybao.objectutilities import TypedList
from .layer import Widget, Layer, TemporaryLayer
from .gridlayer import GridLayer


class LayersManager:

    levels = {
        0:"BACKGROUND",
        1:"MAINGROUND",
        2:"FOREGROUND",
    }
    BACKGROUND = 0
    MAINGROUND = 1
    FOREGROUND = 2
    DEFAULT_LEVEL = MAINGROUND

    def __init__(self, container):

        self._container = container
        self._layers = TypedList(Layer)
        self._default_layer = None
        self._touchable_layers = []

    container = property(lambda self: self._container)
    default_layer = property(lambda self: self._default_layer)
    layers = property(lambda self: self._layers)
    touchable_layers = property(lambda self: self._touchable_layers)

    background = property(lambda self: self.get_level(self.BACKGROUND))
    foreground = property(lambda self: self.get_level(self.FOREGROUND))
    mainground = property(lambda self: self.get_level(self.MAINGROUND))

    def _add_layer(self, layer):
        """
        This method should only be called by Layer.__init__
        """

        assert len(layer) == 0, "A layer should be empty at its creation"
        assert layer.name not in self._layers
        assert layer.level in self.levels

        if layer.touchable:
            if self._default_layer is None:
                self._default_layer = layer
            elif isinstance(self.default_layer, TemporaryLayer) and len(self.default_layer) == 0:
                self.default_layer.kill()
                self._default_layer = layer

        self._layers.append(layer)
        self.sort_layers()
        self._touchable_layers = list(layer for layer in self._layers if layer.touchable)  # this preserv the overlay

        # print("---")
        # for layer in self:
        #     print(layer.name)
        # print("---")

    def _remove_layer(self, layer):
        """
        This method should only be called by Layer.kill
        """

        assert len(layer) == 0
        self._layers.remove(layer)
        if layer.touchable: self._touchable_layers.remove(layer)
        if layer == self.default_layer: self._default_layer = None

    def accept(self, child):
        return isinstance(child, Widget)

    def add(self, child):
        """
        This method should only be called by Container.children._add
        You can override this function in order to define special behaviors
        """

        if self.default_layer is None:  # means there is no layer yet
            self.create_temporary_layer()
            # self._default_layer = TemporaryLayer(self.container, "automaticlyCreatedLayer")

        if child.layer is None:
            self.set_layer_for(child)
        try:
            assert child.layer in self._layers, (child.layer, self.layers)
        except AssertionError:
            print("")
        child.layer.add(child)

    def create_temporary_layer(self, layer_level=None):

        if layer_level is None: layer_level = self.DEFAULT_LEVEL
        for layer in self._layers:
            if layer.level == layer_level and isinstance(layer, TemporaryLayer):
                raise AssertionError("There should only be one TemporaryLayer in the same time")
        return TemporaryLayer(
            self.container, level=layer_level,
            name="automaticlyCreatedLayer{}".format(len(self.layers)))

    def get_ilevel(self, level):

        for layer in self._layers:
            if layer.level == level:
                yield layer

    def get_layer(self, layer_name):

        for layer in self._layers:
            if layer.name == layer_name:
                return layer  # if no layer is found, return None

    def get_level(self, level):

        return tuple(self.get_ilevel(level))

    def remove(self, child):

        assert child.layer in self._layers, layer_name
        child.layer.remove(child)

    def set_default_layer(self, layer):

        if isinstance(layer, str):
            layer = self.get_layer(layer)
        self._default_layer = layer

    def set_layer_for(self, child):

        assert child._layer is None
        if self.default_layer is None:
            self.create_temporary_layer()
        if self.default_layer.accept(child):
            child._layer = self.default_layer
        else:
            for layer in self._layers:
                if layer.accept(child):
                    child._layer = layer
                    break
            if child._layer is None:
                child._layer = self.create_temporary_layer()

    def find_layer_for(self, child, layer_level=None):

        if layer_level is None: layer_level = self.DEFAULT_LEVEL

        if self.default_layer is None:
            self.create_temporary_layer()
        for layer in self.layers:
            if layer.level == layer_level and layer.accept(child):
                return layer
        return self.create_temporary_layer(layer_level)




        if self.default_layer is None:
            return self.create_temporary_layer()
        layer = None
        if self.default_layer.accept(child):
            layer = self.default_layer
        else:
            for layer2 in self._layers:
                if layer2.accept(child):
                    layer = layer2
                    break
            if layer is None:
                layer = self.create_temporary_layer()
        return layer

    def sort_layers(self):

        layer_index = 0
        for level in sorted(self.levels.keys()):
            for layer in sorted(self.get_ilevel(level), key=lambda layer: layer.weight):
                layer._layer_index = layer_index
                layer_index += 1
        self._layers.sort(key=lambda layer: layer.layer_index)

    def swap_layer(self, child, layer):

        assert layer in self._layers, layer
        if child.is_awake:
            assert child in child.layer, child.layer
            child.layer.remove(child)
            layer.add(child)
        child._layer = layer
        child.send_paint_request()





from baopig import *


class UT_Scrollable(Zone):
    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)
        self.set_style_for(Button, height=20)

        scroller = ScrollableZone(
            self, (200, 140), pos=(10, 50), background_color="gray", size=(400, 1000))
        # scroller.set_indicator("Scroller")
        import string
        i = 0
        for h in range(10, 1000, 60):
            b = Button(scroller, pos=(10, h), text=string.ascii_letters[i], padding = 0)
            i += 1
            b.set_indicator("Empty")
            def cut():
                scroller.scrollsliders[0].width -= 20
            if h == 10:
                b.command = cut
            else:
                b.command = lambda: scroller.set_window(scroller.window[:2] + (300, 140))


ut_scenes = [
    UT_Scrollable,
]


if __name__ == "__main__":
    from baopig.unit_tests.TesterScene import TesterScene
    app = Application()
    for scene in ut_scenes:
        TesterScene(app, scene)
    app.launch()






from baopig import *


class UT_Origin_Scene(Scene):

    def __init__(self):
        # TEST : the cross is always at the application center, even after a scene resize
        # TEST : the prisonner (light blue rect) is not visible inside the window (black border on topright corner)
        # TEST : the corners are still at the application corners after a scene resize
        # TEST : the bottomleft corner cannot be dragged, despite the set_dragable
        # TEST : the topright and bottom right corners can be dragged
        # TEST : the topright and bottom right corners follow the scene resizing even after being moved
        # TEST : if the topright corner is set inside the topleft blue border, the prisonner appears
        # TEST : the prisonner can only be seen trought the window
        # TEST : when the topright corner moves, the prisonner is visually static, even with low fps
        # TEST : the prisonner can be dragged
        # TEST : the prisonner can only be seen trought the window, once again
        # TEST : after the scene width changed, the clock abcissa is still at the center of the application
        # TEST : the clock (light gray surface at the center) can be dragged
        # TEST : if the clock has moved, after a scene resizing, it keeps the same distance from the scene's right
        # TEST : the yellow belt (yellow rects around the clock center) cannot be dragged
        # TEST : the yellow center (yellow rect at the clock center) can be dragged
        # TEST : the yellow belt follow the yellow center everywhere he goes
        # TEST : dragging the yellow center don't cause lag
        # TEST : when you drag one of the twins (brown and green rects at the application top), a RecursiveError is raised

        Scene.__init__(self)#, size=(400, 400))

        self.set_mode(pygame.RESIZABLE)

        # CORNERS
        z = Zone(self, origin=Origin(pos=(10, -10), location="bottom", reference_location="bottom", locked=True),
                 size=(100, 100), background_color=(0, 64, 64), name="bottom")
        Dragable.set_dragable(z)
        z = Zone(self, origin=Origin(pos=(-10, -10), location="bottomright", reference_location="bottomright"),
                 size=(100, 100), background_color=(10, 54, 54), name="bottomright")
        Dragable.set_dragable(z)

        # PRISONNER
        z = Zone(self, origin=Origin(pos=(-10, 10), location="right", reference_location="right"),
                 size=(100, 100), background_color=(0, 64, 64), name="right")
        Dragable.set_dragable(z)
        b = Border(self, origin=Origin(pos=(7, 7), reference_comp=self),
                   color=(0, 108, 108), size=(100, 100), width=3, surrounding=True)
        wb = Border(z, origin=Origin(pos=("50%", "50%"), location="center"),
                    size=[z.w-20]*2,
                    surrounding=True, color=(0, 0, 0), width=1)
        r = Rectangle(z, (0, 128, 128), (30, 30), origin=Origin(pos=b.center, location="center", reference_comp=self))
        Dragable.set_dragable(r)
        r.set_window(wb.interior)

        # TODO : tester from_hitbox

        # CLOCK
        z2 = Zone(self, origin=Origin(pos=("-50%", 100), location="midtop", reference_location="topright"),
                  size=(350, 350), background_color=(140, 140, 140), name="z3")
        Dragable.set_dragable(z2)
        z3 = Zone(z2, origin=Origin(pos=(0, 0), location="center", reference_location="center"),
                  size=(250, 250), background_color=(150, 150, 150), name="z2")
        Dragable.set_dragable(z3)
        ref = Rectangle(z3, origin=Origin(pos=("50%", "50%"), location="center", reference_comp=self),
                        color=(128, 128, 0), size=(30, 30), name="ref")
        Dragable.set_dragable(ref)
        import math
        radius = 100
        step = int(math.degrees((2*math.pi)/8))
        for i in range(0, int(math.degrees(2*math.pi)), step):
            r = Rectangle(z2, color=ref.color, size=ref.size,
                        origin=Origin(pos=(math.cos(math.radians(i)) * radius,
                                         math.sin(math.radians(i)) * radius),
                                    reference_comp=ref), name="rect({})".format(i))
        for i in range(0, int(math.degrees(2*math.pi)), step):
            i += step / 2
            r = Rectangle(z3, color=(150, 120, 0), size=ref.size,
                        origin=Origin(pos=(math.cos(math.radians(i)) * radius,
                                         math.sin(math.radians(i)) * radius),
                                    reference_comp=ref), name="rect({})".format(i))

        # TWINS
        r1 = Rectangle(self, pos=("50%", 0),                                    color=(100, 50, 25), size=(30, 30), name="r1")
        r2 = Rectangle(self, origin=Origin(pos=(40, 0), reference_comp=r1), color=(50, 100, 25), size=(30, 30), name="r2")
        r3 = Rectangle(self, origin=Origin(pos=(0, 40), reference_comp=r2),  color=(100, 50, 25), size=(30, 30), name="r3")
        r4 = Rectangle(self, origin=Origin(pos=(-40, 0), reference_comp=r3), color=(50, 100, 25), size=(30, 30), name="r4")
        r1.origin.config(pos=(0, -40), reference_comp=r4)
        Dragable.set_dragable(r1)
        Dragable.set_dragable(r2)
        Dragable.set_dragable(r3)
        Dragable.set_dragable(r4)
        r0 = Rectangle(self, origin=Origin(pos=("50%", "50%"), reference_comp=r1), color=(75, 75, 25), size=(40, 40))
        r0.move_behind(r1)

        # CROSS
        c1 = Rectangle(self, (0, 0, 0), size=(6, 20),
                       origin=Origin(pos=(0, 0), location="center", reference_location="center"))
        c2 = Rectangle(self, (0, 0, 0), size=(20, 6),
                       origin=Origin(pos=(0, 0), location="center", reference_location="center"))


if __name__ == "__main__":
    UT_Origin_Scene()
    launch()




def is_one_of_types(o, *types):
    return isinstance(o, types)
    return True in [isinstance(o, t) for t in types], "{} must be one of {} types, got {} instead".format(o, types, type(o))

def is_iterable(o, k):
    if not hasattr(o, "__iter__"): return False
    return len(o) == k

def is_typed_iterable(o, t, k):
    if not hasattr(o, "__iter__"): return False
    if not len(o) == k: return False
    for elt in o:
        if not isinstance(elt, t): return False
    return True

def is_point(o):
    if not hasattr(o, "__iter__"): return False
    if not len(o) == 2: return False
    for coord in o:
        if not is_one_of_types(coord, int, float): return False
    return True

def is_size(o):
    if not hasattr(o, "__iter__"): return False
    if not len(o) == 2: return False
    for coord in o:
        if not is_one_of_types(coord, int, float): return False
        if not coord >= 0: return False
    return True

def is_color(o):
    if not hasattr(o, "__iter__"): return False
    if not len(o) in (3, 4): return False
    for elt in o:
        if not isinstance(elt, int): return False
        if not 0 <= elt <= 255: return False
    return True



from .scrollablezone import ScrollableZone, ScrollSlider
from .text          import *
from .textedit      import TextEdit
from .lineedit      import LineEdit
from .entry         import Entry  # TODO : remove ? is it abstract ?
from .numentry      import NumEntry
from .label         import *
from .button        import Button, ButtonText
from .checkbox      import CheckBox
from .inputbox      import *
from .progressbar   import ProgressBar
from .slider        import Slider, SliderBloc, SliderBar
from .dialog        import Dialog, DialogFrame, DialogButtonsZone, DialogAnswerButton
from .colorchooserdialog import ColorChooserDialog





import pygame
from baopig.pybao.issomething import *
from baopig.ressources import ressources
from baopig.font.font import _all, Font
from baopig.io import mouse
from baopig._lib import Runable, Widget, Zone, paint_lock
from .text import Text

fonts = _all[ressources.font.file]


class AbstractLabel(Zone):
    """
    A Label is a Container who contain a Text
    The interest of Label is to be able to ask for a given size, and then to adapt the text.
    """

    def __init__(
            self,
            parent,
            text,
            w, h,
            margin=None,  # TODO : better margin
            font=None,
            text_location=None,
            selectable=True,
            **kwargs
    ):
        """

        TODO : documentation a retravailler

        Si le parametre w est renseigne, la taille de la police s'adapte pour que
        le texte ne depasse pas. Sinon, la hitbox s'adapte a la longeur du texte

        Theoriquement parlant, h et font_size sont exactement les memes ; la hauteur
        du label est aussi la taille de la police en pixels. On va d'abord regarder
        si font_size est renseigne, puis h, et on mettra la valeur DEFAULT_FONT_SIZE par defaut.

        margin est l'espace minimal entre les bords de la hitbox et le texte

        Parce que le text_component est ajoute apres la creation du Label, on ne peut deviner
        la taille que prendra ce Label. On reinitialise donc la hitbox et le background dans
        self.set_text_component()
        """

        # DEFAULT VALUES

        # margin, h and font_size
        if font is not None: font = font.copy()
        font_size = font.height if font is not None else None
        font_size_is_unset = font_size is None
        h_is_unset = h is None
        if font_size is not None and h is not None:
            assert font_size <= h, "The text cannot exceed the label height"

        # margin
        if margin is None:
            if font_size is not None and h is not None:
                margin = int((h - font_size) / 2)
            else:
                margin = 0
        assert margin >= 0, "The value of margin must be positive : {} ".format(margin)


        if font_size is None and h is None: font_size = ressources.font.height
        if h is None: h = font_size + margin * 2
        assert h > 0, "The value of h must be positive : {} ".format(h)
        if font_size is None: font_size = h - margin * 2
        assert font_size <= h - margin * 2, "The text cannot exceed the space defined by the margin"
        assert font_size > 0, "The value of font_size must be positive : {} ".format(font_size)

        # w
        if w is None: w = fonts.get_font(font_size).size(text)[0] + margin * 2
        else:
            if font_size_is_unset or True:
                while font_size > 1 and fonts.get_font(font_size).size(text)[0] > w - margin * 2:
                    font_size -= 1
                assert font_size > 0, "The value of font_size is too small : {} " \
                                      "(this is the result of a small width parameter)".format(font_size)
                if h_is_unset and text:
                    while fonts.get_font(font_size).size(text)[0] < w - margin * 2:
                        font_size += 1
                    font_size -= 1
                    h = font_size + margin * 2
            else:
                assert font_size <= w - margin * 2, "The text cannot exceed the space defined by the margin"
        assert w > 0, "The value of w must be positive : {} ".format(w)

        if font is None: font = Font()
        font._height = font_size
        font._color = self.theme.get_value("theme-color-font")  # TODO : remove this temporary fix with better font system

        if text_location is None: text_location = "center"
        assert isinstance(selectable, bool)

        # INHERITANCE
        Zone.__init__(
            self,
            parent=parent,
            size=(w, h),
            **kwargs
        )

        text_component = Text(
            parent=self,
            text=text,
            pos=getattr(self.auto, text_location),
            pos_location=text_location,
            font=font,
            selectable=selectable,
            name=self.name + " -> text"
        )

        # TODO : self.set_padding(margin)

        self._text_component_ref = text_component.get_weakref()
        # self.margin = margin

    text = property(lambda self: self.text_component.text)
    text_component = property(lambda self: self._text_component_ref())
    font_size = property(lambda self: self.text_component.font.height)
    font_color = property(lambda self: self.text_component.font.color)



class Label(AbstractLabel):
    """
    A Label is a Label who cannot evolve during the application life
    It contains a LineText

    """

    def __init__(
            self,
            parent,
            text,
            font=None,
            size=(None, None),
            w=None,
            h=None,
            margin=None,
            text_location=None,
            selectable=True,
            **kwargs
    ):
        """
        Si le parametre w n'est pas renseigne, la hitbox s'adapte a la longeur du texte
        Si le parametre w est renseigne, la taille de la police s'adapte pour que
        le texte ne depasse pas. Si le texte ne prends pas toute la largeur, il est
        positionne au milieu du label

        Theoriquement parlant, h et font_size sont exactement les memes ; la hauteur
        du label est aussi la taille de la police en pixels. On va d'abord regarder
        si font_size est renseigne, puis h, et on mettra la valeur DEFAULT_FONT_SIZE par defaut.

        margin est l'espace minimal entre les bords de la hitbox et le texte
        margin_x est l'espace reel entre les bords droite/gauche de la hitbox et ceux du texte
        margin_y est l'espace reel entre les bords haut/bas de la hitbox et ceux du texte

             ___________________________________________
            |  _ _ _______________________________ _ _  |     | margin = margin_y
            | |   |    _                          |   | |
            |     |   |_   _   _         _        |     |
            | |   |   |   |_| |     |_| |_| |_|   |   | |
            |     |                  _|           |     |
            | | _ |_______________________________| _ | |
            |___________________________________________|     | margin = margin_y

            <---->                                     -
            margin_x                              margin
        """

        assert is_iterable(size, 2)
        if size[0] is not None: w = size[0]
        if size[1] is not None: h = size[1]

        # HERITANCE
        AbstractLabel.__init__(
            self,
            parent,
            text,
            font=font,
            w=w,
            h=h,
            margin=margin,
            text_location=text_location,
            selectable=selectable,
            **kwargs
        )


# TODO : Faire des tests avec des margin, des modification_mode...
class DynamicLabel(AbstractLabel, Runable):
    """
    A DynamicLabel is basically a Label who can change its LineText

    At creation, you must fill one (and only one) of the following parameters :
        - text : The text will not change by itself
        - get_text : At each run() loop, the text will be set to get_text()
        - observed_object : At each run() loop, the text will be set to str(observed_object)

    The aspects of the label who might evolve :
        - the text
        - the font size
        - the font color
        - the background color
        - the margin (TODO : margin modification not implemented yet)

    If the h or w parameter is filled, then they are respectively not able to change at all
    OR
    There is a new modification_mode in addition with these from DynamicText:
        - FIXED_HITBOX : Can change the text. The font size is limited by the hitbox's width
    The default modification_mode is ADAPTABLE

    The surface might wait for the first _render() call to correspond to hitbox size
    This is the only moment where surface.get_size() and hitbox.size can differ

    """
    # TODO : modification_mode rework
    # TODO : pos argument is the position of origin_location

    def __init__(self,
                 parent,
                 pos=None,
                 text=None,
                 get_text=None,
                 observed_object=None,
                 size=(None, None),
                 w=None,
                 h=None,
                 margin=None,
                 font_size=None,
                 font_color=None,
                 background_color=None,
                 selectable=True,
                 modification_mode=None,
                 text_location=None,
                 origin_location=None,
                 name=None,
    ):

        assert len([arg for arg in (text, get_text, observed_object) if arg is not None]) <= 1, \
            "You must fill at most one of these parameters : 'text', 'get_text' or 'observed_object'"

        # text
        if text is None:
            if get_text is None:
                if observed_object is None:
                    text = "Unset"
                else:
                    text = str(observed_object)
                    get_text = observed_object.__str__
            else:
                try:
                    text = str(get_text())
                except Exception as e:
                    text = "Unset"

        # size
        assert is_iterable(size, 2)
        if size[0] is not None: w = size[0]
        if size[1] is not None: h = size[1]

        # modification_mode
        # if modification_mode is None: modification_mode = ADAPTABLE  # TODO : what was that ?

        AbstractLabel.__init__(
            self,
            parent=parent,
            text=text,
            pos=pos,
            w=w,
            h=h,
            margin=margin,
            font_size=font_size,
            font_color=font_color,
            background_color=background_color,
            text_location=text_location,
            origin_location=origin_location,
            name=name,
        )
        Runable.__init__(self)

        self._get_new_text = get_text
        self.modification_mode = modification_mode
        self.has_fixed_width = w is not None or modification_mode == ADAPTABLE
        self.has_fixed_height = h is not None
        # TODO : solve default_font_size bug
        self.default_font_size = self.text_component.font.height if font_size is None else font_size

        if self._get_new_text is None:
            self.stop_running()
        if self.modification_mode == FIXED_HITBOX:
            self.lock_size()
        if self.modification_mode == IMMOVABLE:
            self.lock_size()

    get_new_text = property(lambda self: self._get_new_text)

    def run(self):

        if self.get_new_text is None:
            print("USELESS CALL")
            return
        new_text = self.get_new_text()
        if new_text != self.text:
            self.set_text(new_text)

    def set_font_size(self, font_size):

        assert isinstance(font_size, int)

        if self.text_component.font_size == font_size:
            return None

        if self.modification_mode == IMMOVABLE:
            return None
        elif self.modification_mode == FIXED_FONT_SIZE:
            return None
        elif self.modification_mode == FIXED_TEXT:
            self.text_component.set_font_size(font_size)
        elif self.modification_mode == FIXED_HITBOX:
            while self.w < fonts.get_font(font_size).get_width(self.text):
                font_size -= 1
            self.text_component.set_font_size(font_size)
        elif self.modification_mode == ADAPTABLE:
            self.text_component.set_font_size(font_size)

        self.update_size()
        self.send_paint_request()

    def set_text(self, text):

        assert isinstance(text, str)

        if self.text == text:
            return None

        old_text = self.text

        if self.modification_mode == IMMOVABLE:
            return None
        elif self.modification_mode == FIXED_FONT_SIZE:
            self.text_component.set_text(text)
        elif self.modification_mode == FIXED_TEXT:
            return None
        elif self.modification_mode == FIXED_HITBOX:
            self.text_component.set_text(text)
            self.set_font_size(self.default_font_size)  # the font size need to adapt to the text
        elif self.modification_mode == ADAPTABLE:
            self.text_component.set_text(text)

        self.update_size()
        self.send_paint_request()

    def update_size(self):

        if self.text_component.w + self.margin * 2 != self.w:
            old_w = self.w
            self.resize_width(self.text_component.w + self.margin * 2)
            # self.text_component.x += self.w - old_w

        if self.text_component.h + self.margin * 2 != self.h:
            old_h = self.h
            self.resize_height(self.text_component.h + self.margin * 2)
            # self.text_component.y += self.h - old_h





import pygame
from baopig.pybao.objectutilities import WeakTypedSet, PackedFunctions, PrefilledFunction
from baopig.pybao.issomething import is_point
from baopig.io import mouse, keyboard, clipboard
from .utilities import Linkable, Focusable, paint_lock
from .layer import Layer
from .shapes import Rectangle
from .container import Container


class Selectable():
    """
    A Selectable is a Widget who can be selected

    You are selecting a Selectable when :
        - You click on its selector (a parent), and then move the mouse while it is clicked, and
          the rect representing the drag collide with the Selectable
        - You pressed Ctrl+A while its selector is focused

    The selection_rect closes when :
        - A mouse.LEFTCLICK occurs (not double clicks, wheel clicks...)
        - The Selectable size or position changes

    The default reaction is to call handle_select() and handle_unselect()
    You can rewrite the check_select(abs_rect) method for accurate selections (like SelectableText)
    You can rewrite the select() and unselect() methods for specific behaviors (like SelectableText)

    The get_selected_data() method return, when the Selectable is selected, the Selectable itself
    You will probably want to override this method
    """

    def __init__(self):

        self._is_selected = False

        selector = self.parent
        while not isinstance(selector, Selector):
            selector = selector.parent
        self._selector_ref = selector.get_weakref()

        self.handle_select = PackedFunctions()
        self.handle_unselect = PackedFunctions()

        # self.handle_select.add(PrefilledFunction(print, "HANDLE SELECT :", self))
        # self.handle_unselect.add(PrefilledFunction(print, "HANDLE UNSELECT :", self))

        if hasattr(self.selector, "selectables"):  # selector might not have been initialized
            self.selector.selectables.add(self)

        self.connect("unselect", self.signal.NEW_SURF)
        self.connect("unselect", self.signal.MOTION)
        self.connect("unselect", self.signal.HIDE)

    is_selected = property(lambda self: self._is_selected)
    selector = property(lambda self: self._selector_ref())

    def check_select(self, selection_rect):
        """
        Method called by the selector each time the selection_rect rect changes
        The selection_rect has 3 attributes:
            - start
            - end
            - rect (the surface between start and end)
        These attributes are absolute referenced, wich means they are relative
        to the application. Start and end attributes reffer to the start and end
        of the selection_rect, who will often be caused by a mouse link motion
        """

        assert self.get_weakref()._comp is not None
        if self.abs_hitbox.colliderect(selection_rect.abs_hitbox):
            if not self.is_selected: self.handle_select()  # TODO : if not used, remove it ?
            self._is_selected = True
            self.select()
        else:
            if not self.is_selected: return
            self._is_selected = False
            self.unselect()
            self.handle_unselect()

    def get_selected_data(self):

        if self.is_selected:
            return self

    # TODO : decorate
    def select(self):
        """
        Called each time the selection rect move and collide with this Selectable
        """

    def unselect(self):
        """
        Called when the selection rect don't collide anymore with this Selectable
        """


class DefaultSelectionRect(Rectangle):
    """
    A SelectionRect is a rect relative to the application
    """

    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        color = "theme-color-selection_rect",
        border_color = "theme-color-selection_rect_border",
        border_width = 1,
    )

    def __init__(self, parent, start, end):

        Rectangle.__init__(
            self,
            parent=parent,
            pos=(0, 0),
            size=(0, 0),
            layer=parent.selectionrect_layer,
            name=parent.name+".selection_rect"
        )
        self.start = None
        self.end = None
        self.parent._selection_rect_ref = self.get_weakref()
        self.set_visibility(parent._selectionrect_visibility)
        self.connect("hide", self.parent.signal.UNLINK)

        self.set_start(start)
        self.set_end(end)

    def set_end(self, abs_pos):
        assert self.start is not None
        assert is_point(abs_pos)
        self.end = abs_pos
        left = min((self.start[0], self.end[0]))
        top = min((self.start[1], self.end[1]))
        rect = pygame.Rect(self.start,
                           (self.end[0] - self.start[0],
                            self.end[1] - self.start[1]))
        rect.normalize()
        # rect = self.parent.abs_hitbox.clip(rect)
        self.move_at(self.reference(rect.topleft))
        self.resize(rect.w+1, rect.h+1)  # the selection_rect rect collide with mouse.pos
        self.clip(self.parent.auto_hitbox)

    def set_start(self, abs_pos):
        assert self.end is None
        assert is_point(abs_pos)
        self.start = abs_pos


class Selector(Linkable):
    """
    Abstract class for components who need to handle when they are linked
    and then, while the mouse is still pressed, to handle the mouse drag in
    order to simulate a rect from the link origin to the link position and
    select every Selectable object who collide with this rect
    """

    def __init__(self, SelectionRectClass=None):

        # TODO : test : a disabled Selector cannot select

        if SelectionRectClass is None: SelectionRectClass = DefaultSelectionRect

        assert isinstance(self, Container)
        assert issubclass(SelectionRectClass, DefaultSelectionRect)
        Linkable.__init__(self)

        class Selectables(WeakTypedSet):
            def __init__(selectables):
                WeakTypedSet.__init__(selectables, ItemsClass=Selectable)
                def add_selectables(cont):
                    for comp in cont.children:
                        if hasattr(comp, "children"):
                            add_selectables(comp)
                        if isinstance(comp, Selectable):
                            selectables.add(comp)
                add_selectables(self)
            def add(selectables, comp):
                if comp not in selectables:
                    super().add(comp)
                    comp.signal.KILL.connect(lambda: selectables.remove(comp))
            def get_selected(selectables):
                for comp in selectables:
                    if comp.is_selected:
                        yield comp
            selected = property(get_selected)
        self.selectables = Selectables()

        self._can_select = False
        self._selection_rect_ref = lambda: None
        self._selection_rect_class = SelectionRectClass
        self._selectionrect_visibility = True
        self.selectionrect_layer = None

        self.connect("_interact_with_clipboard", self.signal.KEYDOWN)
        self.connect("close_selection", self.signal.DEFOCUS)
        self.connect("close_selection", self.signal.LINK)  # only usefull at link while focused

    is_selecting = property(lambda self: self._selection_rect_ref() is not None)
    selection_rect = property(lambda self: self._selection_rect_ref())

    def _interact_with_clipboard(self, key):

        if keyboard.mod.cmd:
            if key is pygame.K_a:
            # Cmd + a -> select all
                self.select_all()
            elif key is pygame.K_c:
            # Cmd + c -> copy selected data
                if self.is_selecting:
                    clipboard.put(self.get_selected_data())  # TODO : fix get_selected_data
            elif key is pygame.K_v:
            # Cmd + v -> paste clipboard data
                self.paste(clipboard.get(str) if clipboard.contains(str) else '')
            elif key is pygame.K_x:
            # Cmd + x -> copy and cut selected data
                if self.is_selecting:
                    clipboard.put(self.get_selected_data())
                    self.del_selected_data()

    def close_selection(self):

        if not self.is_selecting: return
        self.selection_rect.kill()
        for selectable in self.selectables.selected:
            if not selectable.is_selected: continue
            selectable._is_selected = False
            selectable.unselect()
            selectable.handle_unselect()

    def copy(self):
        """
        Copy the selected data in clipboard
        """
        data = self.get_selected_data()
        if data:
            clipboard.put(data)

    def cut(self):
        """
        To be overriden
        """

    def enable_selecting(self, can_select=True):
        """
        Completely delete the ability to make selections in a Selector
        """

        if bool(can_select) == self._can_select: return
        self._can_select = bool(can_select)
        if can_select:
            self.selectionrect_layer = Layer(self, self._selection_rect_class, name="selectionrect_layer",
                                             level=self.layers_manager.FOREGROUND, maxlen=1, touchable=False)
        else:
            self.close_selection()
            self.selectionrect_layer.kill()
            self.selectionrect_layer = None

    def end_selection(self, abs_pos, visible=None):
        """
        :param abs_pos: An absolute position -> relative to the scene
        :param visible: If you want to change the visibility until the next end_selection
        """

        if not self._can_select: return
        assert self.selection_rect is not None
        if abs_pos == self.selection_rect.end: return
        if visible is not None: self.selection_rect.set_visibility(visible)
        else:                   self.selection_rect.set_visibility(self._selectionrect_visibility)
        self.selection_rect.set_end((abs_pos[0], abs_pos[1]))
        for selectable in self.selectables:
            selectable.check_select(self.selection_rect)

    def get_selected_data(self):

        if not self.is_selecting: return None
        selected = tuple(s for s in self.selectables if s.is_selected)
        if not selected: return
        sorted_selected = sorted(selected, key=lambda o: (o.abs.top, o.abs.left))
        return '\n'.join(str(s.get_selected_data()) for s in sorted_selected)

    def handle_link_motion(self, link_motion_event):
        with paint_lock:
            if self.selection_rect is None:
                self.start_selection(link_motion_event.origin)
            self.end_selection(link_motion_event.pos)

    def paste(self, data):
        """
        This method is called when the user press Cmd + V
        data is the string content of the clipboard
        If the clipboard don't contains string data, then data is ''
        If you want to acces to other types of clipboard data, try :
        if clipboard.contains(type):
            data = clipboard.get(type)
        """

    def select_all(self):

        if self.selectables:
            if self.is_selecting:
                self.close_selection()
            self.start_selection(self.abs.topleft)
            self.end_selection(self.abs.bottomright, visible=False)

            # self.start_selection((
            #     min(s.abs.left for s in self.selectables),
            #     min(s.abs.top for s in self.selectables)
            # ))
            # self.end_selection((
            #     max(s.abs.right for s in self.selectables),
            #     max(s.abs.bottom for s in self.selectables)
            # ))

    def set_selectionrect_visibility(self, visible):

        self._selectionrect_visibility = bool(visible)
        if self.selection_rect is not None:
            self.selection_rect.set_visibility(visible)

    def start_selection(self, abs_pos):
        """
        A selection_rect can only be started once
        :param abs_pos: An absolute position -> relative to the scene
        """
        if not self._can_select: return
        if self.selection_rect is not None:
            raise PermissionError("A selection must be closed before creating a new one")
        self._selection_rect_class(self, abs_pos, abs_pos)




from baopig.pybao.objectutilities import WeakTypedList
from .widget import Widget, Communicative


class Layer(Communicative):
    """
    A Layer is a manager who contains some of its container's children
    Every component is stored in one of its parent's layers
    The positions of components inside a layer define the overlay : first behind, last in front.
    Eache layer can be overlay in the foreground, the main ground or the background.
    Layers from one ground are overlay depending on their weight : a weight of 0 means it need
    to stand behind a layer with weight 6. The default weight is 2.
    """

    def __init__(self, container, *filter, name=None, level=None, weight=None,
                 default_sortkey=None, sort_by_pos=False, touchable=True, maxlen=None, adaptable=False):
        """
        :param container: the Container who owns the layer
        :param name: a unic identifier for the layer
        :param filter: a class or list of class from wich every layer's component must herit
        :param level: inside the container's layers : lowest behind greatest in front, default to MAINGROUND
        :param weight: inside the layer's level : lowest behind greatest in front, default to 2
        :param default_sortkey: default key fo layer.sort(). if set, at each append, the layer will be sorted
        :param sort_by_pos: if set, the default sortkey will be a function who sort components by y then x
        :param touchable: components of non touchable layer are not hoverable
        :param maxlen: the maximum numbers of components the layer can contain
        """

        if name is None: name = "UnnamedLayer{}".format(len(container.layers))
        if not filter: filter = [Widget]
        if weight is None: weight = 2
        if level is None: level = container.layers_manager.DEFAULT_LEVEL
        if sort_by_pos:
            assert default_sortkey is None
            default_sortkey = lambda c: (c.top, c.left)

        assert isinstance(name, str), name
        assert name not in container.layers, name
        for filter_class in filter: assert issubclass(filter_class, Widget), filter_class
        if default_sortkey is not None: assert callable(default_sortkey), default_sortkey
        if maxlen is not None: assert isinstance(maxlen, int), maxlen
        assert isinstance(weight, (int, float)), weight
        assert level in container.layers_manager.levels, level

        Communicative.__init__(self)

        # NOTE : adaptable, container, name, touchable and level are not editable, because you
        #        need to know what kind of layer you want since its creation
        self._is_adaptable = bool(adaptable)
        self._comps = WeakTypedList(*filter)
        self._container = container
        self._filter = filter
        self.default_sortkey = default_sortkey  # don't need protection
        self._layer_index = None  # defined by container.layers
        self._layers_manager = container.layers_manager
        self._maxlen = maxlen
        self._name = name
        self._touchable = bool(touchable)
        self._level = level
        self._weight = weight

        self.layers_manager._add_layer(self)

    def __add__(self, other):
        return self._comps + other

    def __bool__(self):
        return bool(self._comps)

    def __contains__(self, item):
        return self._comps.__contains__(item)

    def __getitem__(self, item):
        return self._comps.__getitem__(item)

    def __iter__(self):
        return self._comps.__iter__()

    def __len__(self):
        return self._comps.__len__()

    def __repr__(self):
        return "{}(name:{}, index:{}, filter:{}, touchable:{}, level:{}, weight:{}, components:{})".format(
            # "Widgets" if self.touchable else "",
            self.__class__.__name__, self.name, self._layer_index, self._filter, self.touchable,
            self.level, self.weight, self._comps)

    is_adaptable = property(lambda self: self._is_adaptable)
    container = property(lambda self: self._container)
    layer_index = property(lambda self: self._layer_index)
    layers_manager = property(lambda self: self._layers_manager)
    level = property(lambda self: self._level)
    maxlen = property(lambda self: self._maxlen)
    name = property(lambda self: self._name)
    touchable = property(lambda self: self._touchable)
    weight = property(lambda self: self._weight)

    def accept(self, comp):

        if self.maxlen and self.maxlen <= len(self._comps): return False
        return self._comps.accept(comp)

    def add(self, comp):
        """
        WARNING : This method should only be called by the LayersManager: cont.layers_manager
        You can override this function in order to define special behaviors
        """
        if self.maxlen and self.maxlen <= len(self._comps):
            raise PermissionError("The layer is full (maxlen:{})".format(self.maxlen))

        self._comps.append(comp)
        if self.default_sortkey:
            self.sort()

        if self.is_adaptable:
            self.container.adapt(self)

    def clear(self):

        for comp in tuple(self._comps):
            comp.kill()

    def _find_place_for(self, comp):

        assert self.accept(comp), f"{self} don't accept {comp}"
        return (0, 0)

    def get_visible_comps(self):
        for comp in self._comps:
            if comp.is_visible:
                yield comp
    visible = property(get_visible_comps)

    def index(self, comp):
        return self._comps.index(comp)

    def kill(self):

        self.clear()
        self.layers_manager._remove_layer(self)

    def move_comp1_behind_comp2(self, comp1, comp2):
        assert comp1 in self._comps, "{} not in {}".format(comp1, self)
        assert comp2 in self._comps, "{} not in {}".format(comp2, self)
        self.overlay(self.index(comp2), comp1)

    def move_comp1_in_front_of_comp2(self, comp1, comp2):
        assert comp1 in self, "{} not in {}".format(comp1, self)
        assert comp2 in self, "{} not in {}".format(comp2, self)
        self.overlay(self.index(comp2) + 1, comp1)
        # self._remove(comp1)
        # super().insert(self.index(comp2) + 1, comp1)
        # self._warn_change(comp1.hitbox)

    def overlay(self, index, comp):
        """
        Move a component at index
        """

        assert comp in self._comps
        self._comps.remove(comp)
        self._comps.insert(index, comp)
        self.container._warn_change(comp.hitbox)

    def pack_TO_MOVE_OUT(self, key=None, horizontal=False, margin=None, padding=None, adapt=True):
        """
        Place children on one row or one column, sorted by key (default : pos)
        If horizontal is False, they will be placed vertically
        The margin is the space between two children
        The margin is the minimum space between a comp and the container's border
        If adapt is True, the container's sizes will try to adapt to the layer's
        children, including the padding
        """
        if key is None: key = lambda o: (o.top, o.left)
        if margin is None: margin = 0
        if padding is None: padding = 0

        sorted_children = sorted(self, key=key)

        if horizontal:
            left = padding
            for comp in self:
                comp.topleft = (left, padding)
                if comp.topleft != (left, padding):
                    raise PermissionError("Cannot pack a layer who contains locked children")
                left = comp.right + margin
        else:
            top = padding
            for comp in self:
                comp.topleft = (padding, top)
                if comp.topleft != (padding, top):
                    raise PermissionError("Cannot pack a layer who contains locked children")
                top = comp.bottom + margin

        if adapt:
            self.container.adapt(self, padding=padding)

    def remove(self, comp):
        """
        WARNING : This method should only be called by the LayersManager: cont.layers_manager
        You can override this function in order to define special behaviors
        """
        self._comps.remove(comp)

        if self.is_adaptable:  # TODO : test to adapt a Scene at (0, 0)
            self.container.adapt(self)

    def set_filter(self, filter):

        self._comps.set_ItemsClass(filter)

    def set_maxlen(self, maxlen):

        assert isinstance(maxlen, int) and len(self._comps) <= maxlen
        self._maxlen = maxlen

    def set_weight(self, weight):

        assert isinstance(weight, (int, float))
        self._weight = weight
        self.layers_manager.sort_layers()

    def sort(self, key=None):
        """
        Permet de trier les enfants d'un layer selon une key
        Cette fonction ne deplace pas les enfants, elle ne fait que changer leur
        superpositionnement
        """
        if key is None: key = self.default_sortkey
        self._comps.sort(key=key)


class TemporaryLayer(Layer):
    """
    A TemporaryLayer will kill itself when its last component is removed
    """

    def __init__(self, *args, **kwargs):

        Layer.__init__(self, *args, **kwargs)

    def remove(self, comp):

        super().remove(comp)
        if len(self._comps) is 0:
            for comp in self.container.children.sleeping:
                if comp.layer is self:
                    return
            self.kill()





import time
from pygame import Rect
from baopig.io import mouse
from .hoverable import Hoverable
from .container import Container


class Scrollable(Hoverable):
    """If the widget is hovered and the mouse is scrolling, scroll

    scrollaxis is one of '', 'x', 'y', 'xy'

    The scroll is between window and rect
    You should write
        scrollable_widget.set_window(something)

    """

    def __init__(self, scrollaxis=None):

        assert isinstance(self, Container)
        if scrollaxis is None:
            scrollaxis = "y"
        assert scrollaxis in ("", "x", "y", "xy")
        self._scrollaxis = scrollaxis

    scrollaxis = property(lambda self: self._scrollaxis)

    def scroll(self, dx, dy):

        assert self.window is not None
        window = Rect(self.window)

        if dx < 0:
            right = self.right + dx
            if window.right > right:
                dx = window.right - self.right
        elif dx > 0:
            left = self.left + dx
            if window.left < left:
                dx = window.left - self.left

        if dy < 0:
            bottom = self.bottom + dy
            if window.bottom > bottom:
                dy = window.bottom - self.bottom
        elif dy > 0:
            top = self.top + dy
            if window.top < top:
                dy = window.top - self.top

        if (dx == 0) and (dy == 0):
            return

        self.move(dx, dy)


class ScrollableByMouse(Scrollable, Hoverable):

    def __init__(self, scrollaxis=None):

        Scrollable.__init__(self, scrollaxis)
        Hoverable.__init__(self)

        self._last_scroll_time = time.time()  # for faster scroll when together
        self.connect("handle_mouse_scroll", mouse.signal.SCROLL)

    def handle_mouse_scroll(self, scroll_event):
        """
        WARNING : this is not properly coded.
        If this widget contains a Scrollable widget, both can scroll together
        If a widget is hovering this one, it still can be scrolled
        """

        if self.window is None: return
        if not self.collidemouse(): return
        accelerator = 1
        old_scroll_time = self._last_scroll_time
        self._last_scroll_time = time.time()
        d = self._last_scroll_time - old_scroll_time

        if d < 0.5:
            accelerator = 20
        elif d < 0.1:
            accelerator = 5

        if "y" in self.scrollaxis:
            self.scroll(0, scroll_event.direction * accelerator)
        else:
            self.scroll(scroll_event.direction * accelerator, 0)




# TODO : replace sticky and pos_ref and pos_ref_location by flags ?


import inspect
import pygame
from baopig.pybao.issomething import *
from baopig.pybao.objectutilities import EditableTuple, Object, PackedFunctions, TypedSet
from baopig._debug import *
from baopig.io import LOGGER, mouse
from baopig.communicative import Communicative
from .utilities import paint_lock, functools
from .style import HasStyle, StyleClass, Theme


if debug_del:
    _is_alive = True
    import gc  # garbage collector


class MetaPaintLocker(type):
    def __call__(self, *args, **kwargs):
        with paint_lock:
            widget = super().__call__(*args, **kwargs)
            if widget.collidemouse():
                mouse.update_hovered_comp()
            return widget


class MetaUndefinedType(type):
    def __instancecheck__(self, instance):
        return True


class UndefinedType(metaclass=MetaUndefinedType):
    """This class is used for undefined style types, since isinstance(obj, UndefinedType) -> True"""
    def __init__(self):
        raise PermissionError("This class should never be instanciated")


class WeakRef:
    def __init__(self, comp):
        self._comp = comp

    def __call__(self):
        return self._comp


class _Origin:
    """
    An Origin is referenced by its parent
    When the parent moves, the component follows

    A Origin coordinate can be given in different ways:
        - x = 4         sets x to 4 pixels
        - x = '10%'     sets x to self.parent.width * 10 / 100 (automatically updated)

    WARNING : A component with a dynamic origin (pourcentage position) cannot be manually
              moved by any other way than redefining the origin
    """

    def __init__(self, owner):

        """
        asked_pos is the distance between reference at reference_location and owner at location
        """
        self._owner_ref = owner.get_weakref()
        self._asked_pos = owner.style["pos"]
        self._from_hitbox = owner.style["pos_from_ref_hitbox"]
        self._location = owner.style["pos_location"]
        try:
            self._has_adaptable_size = isinstance(self.owner.style["width"], str) or isinstance(self.owner.style["height"], str)
        except KeyError:
            self._has_adaptable_size = False
        # We need to initialize reference before config
        reference = owner.style["pos_ref"]
        if reference is None: reference = owner.parent
        self._reference_ref = reference.get_weakref()
        self.reference.signal.RESIZE.connect(self._weak_update_owner_pos, owner=self.owner)
        if self._has_adaptable_size:
            self.reference.signal.RESIZE.connect(self.owner._update_size, owner=self.owner)
        if self.reference != owner.parent:
            self.reference.signal.MOTION.connect(self._weak_update_owner_pos, owner=self.owner)
        self._reference_location = owner.style["pos_ref_location"]
        try:
            assert self._reference_location != "top"
        except AssertionError:
            raise AssertionError

    def __str__(self):
        return "Origin(asked_pos={}, location={}, reference={}, reference_location={}, from_hitbox={}, " \
               "is_locked={})".format(self.asked_pos, self.location, self.reference,
                         self.reference_location, self.from_hitbox, self.owner.has_locked.origin)

    asked_pos = property(lambda self: self._asked_pos)
    from_hitbox = property(lambda self: self._from_hitbox)
    is_locked = property(lambda self: self.owner.has_locked.origin)
    location = property(lambda self: self._location)
    owner = property(lambda self: self._owner_ref())
    reference = property(lambda self: self._reference_ref())
    reference_location = property(lambda self: self._reference_location)

    def _reset_asked_pos(self):
        """
        This method should only be called by Widget._move()
        If the asked_pos was including percentage, and the percentage don't match anymore, it will be
        replaced by an integer value
        """

        old_pos = self.pos
        current_pos = getattr(self.owner.rect, self.location)
        if old_pos == current_pos:
            return

        if self.is_locked:
            raise AssertionError(self)

        owner_abspos_at_location = getattr(self.owner.abs, self.location)
        reference_abspos_at_location = getattr(self.reference.abs, self.reference_location)

        self._asked_pos = (
            owner_abspos_at_location[0] - reference_abspos_at_location[0],
            owner_abspos_at_location[1] - reference_abspos_at_location[1]
        )

    def _weak_update_owner_pos(self):
        new_pos = self.pos
        old_pos = getattr(self.owner.rect, self.location)
        if new_pos == old_pos: return
        self.owner._move(dx=new_pos[0] - old_pos[0], dy=new_pos[1] - old_pos[1])

    @staticmethod
    def accept(coord):

        if isinstance(coord, str):
            if not coord.endswith('%'):
                return False
            try:
                int(coord[:-1])
                return True
                return 0 <= int(coord[:-1]) <= 100
            except ValueError:
                return False

        elif hasattr(coord, "__iter__"):
            if len(coord) != 2: return False
            return False not in (_Origin.accept(c) for c in coord)

        try:
            coord = int(coord)
            return True
        except ValueError:
            raise TypeError("Wrong position value : {} (see documentation above)".format(coord))
            return False

    def config(self, pos=None, location=None, reference_comp=None, reference_location=None,
               from_hitbox=None, locked=None):

        old_pos = self.pos if self._asked_pos else None
        old_location = self.location

        if locked is False:
            self.owner.has_locked.origin = False

        if self.is_locked:
            raise PermissionError("This origin is locked")

        if location is not None:
            self._location = WidgetLocation(location)

        if reference_comp is not None:
            assert isinstance(reference_comp, Widget)
            self.reference.signal.MOTION.rem_command(self._weak_update_owner_pos)
            self.reference.signal.RESIZE.rem_command(self._weak_update_owner_pos)
            if self._has_adaptable_size:
                self.reference.signal.RESIZE.rem_command(self.owner._update_size)
            self._reference_ref = reference_comp.get_weakref()
            self.reference.signal.RESIZE.connect(self._weak_update_owner_pos, owner=self.owner)
            if self.reference != self.owner.parent:
                self.reference.signal.MOTION.connect(self._weak_update_owner_pos, owner=self.owner)
            if self._has_adaptable_size:
                self.reference.signal.RESIZE.connect(self.owner._update_size)

        if reference_location is not None:
            self._reference_location = WidgetLocation(reference_location)

        if from_hitbox is not None:
            self._from_hitbox = bool(from_hitbox)

        if pos is not None:
            assert _Origin.accept(pos), f"Wrong position value : {pos} (see documentation above)"
            self._asked_pos = pos

        self.owner.move_at(self.pos, self.location)

        if locked is True:
            self.owner.has_locked.origin = True

    def get_pos_relative_to_owner_parent(self):

        pos = []

        if self.from_hitbox:  # from the reference hitbox

            # Percentage translation of asked_pos
            for i, c in enumerate(self._asked_pos):

                if isinstance(c, str):
                    if debug_with_assert:
                        if not c.endswith('%'):
                            raise ValueError(f"Uncorrect coordinate : {c}")
                    c = self.reference.hitbox.size[i] * int(c[:-1]) / 100

                pos.append(int(c))

            # Transition at reference_location
            if self._reference_location != "topleft":
                d = getattr(self.reference.auto_hitbox, self._reference_location)
                for i in range(2):
                    pos[i] += d[i]

            # Set pos relative to self.owner.parent
            if self.reference is not self.owner.parent:
                pos = (
                    pos[0] + self.reference.abs_hitbox.left - self.owner.parent.abs.left,
                    pos[1] + self.reference.abs_hitbox.top - self.owner.parent.abs.top
                )

        else:
            # Percentage translation of asked_pos
            for i, c in enumerate(self._asked_pos):

                if isinstance(c, str):
                    if debug_with_assert:
                        if not c.endswith('%'):
                            raise ValueError(f"Uncorrect coordinate : {c}")
                    c = self.reference.rect.size[i] * int(c[:-1]) / 100

                pos.append(int(c))

            # Transition at reference_location
            if self._reference_location != "topleft":
                dx, dy = getattr(self.reference.auto_rect, self._reference_location)
                pos = pos[0] + dx, pos[1] + dy

            # Set pos relative to self.owner.parent
            if self.reference != self.owner.parent:
                pos = (
                    pos[0] + self.reference.abs.left - self.owner.parent.abs.left,
                    pos[1] + self.reference.abs.top - self.owner.parent.abs.top
                )

        return tuple(pos)
    pos = property(get_pos_relative_to_owner_parent)

    def lock(self):

        self.owner.has_locked.origin = True

    def unlock(self):

        self.owner.has_locked.origin = False


class _Window(tuple):
    follow_movements = False


class WidgetLocation(str):

    ACCEPTED = ("topleft", "midtop", "topright",
                "midleft", "center", "midright",
                "bottomleft", "midbottom", "bottomright")

    def __new__(cls, location):

        if location == "left":
            location = "midleft"
        elif location == "top":
            location = "midtop"
        elif location == "right":
            location = "midright"
        elif location == "bottom":
            location = "midbottom"
        if location not in WidgetLocation.ACCEPTED:
            raise ValueError(f"Wrong value for location : '{location}', "
                             f"must be one of {WidgetLocation.ACCEPTED}")
        return str.__new__(cls, location)


class ProtectedHitbox(pygame.Rect):
    """
    A ProtectedHitbox cannot be resized or moved
    """

    def __init__(self, *args, **kwargs):
        pygame.Rect.__init__(self, *args, **kwargs)
        object.__setattr__(self, "_mask", None)
        object.__setattr__(self, "err", PermissionError("A ProtectedHitbox cannot change at all"))

    def __setattr__(self, *args):

        raise PermissionError

    pos = property(lambda self: tuple(self[:2]))
    lpos = property(lambda self: self[:2], doc="A list containing the rect topleft")

    def clamp_ip(self, Rect):

        raise self.err

    def colliderect(self, rect):
        # TODO : collidemask
        return super().colliderect(rect)

    def inflate_ip(self, x, y):

        raise self.err

    def move_ip(self, x, y):

        raise self.err

    def referencing(self, pos):
        """
        Return a new pos, relative the hitbox topleft
        The pos is considered as relative to (0, 0)

        Example :
            hibtox = ProtectedHitbox(100, 200, 30, 30)
            pos = (105, 205)
            hitbox.referencing(pos) -> (5, 5)
        """
        return pos[0] - self.left, pos[1] - self.top

    def set_mask(self):
        # TODO : use cases
        raise PermissionError("not implemented yet")

    def union_ip(self, Rect):

        raise self.err

    def unionall_ip(self, Rect_sequence):

        raise self.err


class HasProtectedHitbox:
    """
    The pos parameter is in fact the origin, who will often coincide with the topleft
    """

    def __init__(self):

        """
        rect is the surface hitbox, relative to the parent
        abs_rect is the rect relative to the application (also called 'abs')
        auto_rect is the rect relative to the component itself -> topleft = (0, 0) (also called 'auto')
        window is the rect inside wich the surface can be seen, relative to the parent
        hitbox is the result of clipping rect and window.
        If window is set to None, the hitbox will equal to the rect
        abs_hitbox is the hitbox relative to the application
        auto_hitbox is the hitbox relative to the component itself
        """

        size = self.surface.get_size()
        self._rect = ProtectedHitbox((0, 0), size)
        self._abs_rect = ProtectedHitbox((0, 0), size)
        self._auto_rect = ProtectedHitbox((0, 0), size)
        self._window = None
        self._hitbox = ProtectedHitbox((0, 0), size)
        self._abs_hitbox = ProtectedHitbox((0, 0), size)
        self._auto_hitbox = ProtectedHitbox((0, 0), size)

        """
        MOTION is emitted when the absolute position of the component.rect moves
        It sends two parameters : dx and dy
        WARNING : the dx and dy are the motion of the component.rect ! This means, when the
                  component's parent moves, a MOTION.emit(dx=0, dy=0) is probalby raised.
                  In facts, if the component.origin.reference is not the parent, the motion
                  won't be dx=0, dy=0
        NOTE : an hitbox cannot move and be resized in the same time
        """
        self.create_signal("MOTION")
        self.create_signal("NEW_WINDOW")

        """
        A margin is a defined rectangle around the component's surface
        A margin have left, top, right and bottom attributes, they are the amount of
        space required between the hitbox sides and the surface
        The hitbox contain the margin, so it is occasionnaly bigger than surface
        """
        self._margin_TBR = Object(
            left=0,
            top=0,
            right=0,
            bottom=0,
        )

        """
        Change a surface via set_surface(...) is the only way to change a cmponent size
        Changing size will emit self.signal.RESIZE if the component is visible
        A component with locked size cannot be resized, but the surface can change
        
        When a component is resized, we need a point of reference (called origin) whose pixel
        will not move
        
        origin.location can be one of : topleft,      midtop,     topright,
                                        midleft,      center,     midright,
                                        bottomleft,   midbottom,  bottomright
        
        Exemple : origin.location = midright
                  if the new size is 10 more pixel on the width, then the origin is on the right,
                  so we add the 10 new pixels to the left 
        """
        self.create_signal("RESIZE")

        # This will initialize the rects and hiboxes
        self._origin = _Origin(owner=self)


        # self.move_at(self.origin.pos, self.origin.location)

        new = self.origin.pos
        old = getattr(self.rect, self.origin.location)
        # Rewrite a part of the move_at() method so we can skip this line
        # if old == new: return
        self._move(new[0] - old[0], new[1] - old[1])

        # print("NEW COMPONENT :", repr(self))

    # GETTERS ON PROTECTED FIELDS
    margin_TBR =            property(lambda self: self._margin)
    origin =            property(lambda self: self._origin)

    # HITBOX
    rect =              property(lambda self: self._rect)
    abs = abs_rect =    property(lambda self: self._abs_rect)
    auto = auto_rect =  property(lambda self: self._auto_rect)
    window =            property(lambda self: self._window)
    hitbox =            property(lambda self: self._hitbox)
    abs_hitbox =        property(lambda self: self._abs_hitbox)
    auto_hitbox =       property(lambda self: self._auto_hitbox)

    bottom =            property(lambda self: self._rect.bottom,        lambda self, v: self.move_at(v, "bottom"))
    bottomleft =        property(lambda self: self._rect.bottomleft,    lambda self, v: self.move_at(v, "bottomleft"))
    bottomright =       property(lambda self: self._rect.bottomright,   lambda self, v: self.move_at(v, "bottomright"))
    center =            property(lambda self: self._rect.center,        lambda self, v: self.move_at(v, "center"))
    centerx =           property(lambda self: self._rect.centerx,       lambda self, v: self.move_at(v, "centerx"))
    centery =           property(lambda self: self._rect.centery,       lambda self, v: self.move_at(v, "centery"))
    left =              property(lambda self: self._rect.left,          lambda self, v: self.move_at(v, "left"))
    midbottom =         property(lambda self: self._rect.midbottom,     lambda self, v: self.move_at(v, "midbottom"))
    midleft =           property(lambda self: self._rect.midleft,       lambda self, v: self.move_at(v, "midleft"))
    midright =          property(lambda self: self._rect.midright,      lambda self, v: self.move_at(v, "midright"))
    midtop =            property(lambda self: self._rect.midtop,        lambda self, v: self.move_at(v, "midtop"))
    pos = topleft =     property(lambda self: self._rect.topleft,       lambda self, v: self.move_at(v, "topleft"))
    right =             property(lambda self: self._rect.right,         lambda self, v: self.move_at(v, "right"))
    top =               property(lambda self: self._rect.top,           lambda self, v: self.move_at(v, "top"))
    topright =          property(lambda self: self._rect.topright,      lambda self, v: self.move_at(v, "topright"))
    x =                 property(lambda self: self._rect.x,             lambda self, v: self.move_at(v, "x"))
    y =                 property(lambda self: self._rect.y,             lambda self, v: self.move_at(v, "y"))

    h =                 property(lambda self: self._rect.h)
    height =            property(lambda self: self._rect.height)
    size =              property(lambda self: self._rect.size)
    w =                 property(lambda self: self._rect.w)
    width =             property(lambda self: self._rect.width)

    def _move(self, dx, dy):

        old_hitbox = tuple(self.hitbox)
        with paint_lock:

            pygame.Rect.__setattr__(self.rect, "left", self.rect.left+dx)
            pygame.Rect.__setattr__(self.rect, "top", self.rect.top+dy)
            pygame.Rect.__setattr__(self.abs_rect, "topleft", (self.parent.abs.left + self.left, self.parent.abs.top + self.top))

            if self.window is not None:

                if self.window.follow_movements:
                    self._window = _Window((self.window[0] + dx, self.window[1] + dy) + self.window[2:])
                    self.window.follow_movements = True
                    pygame.Rect.__setattr__(self.hitbox, "topleft", (self.hitbox.left+dx, self.hitbox.top+dy))
                    pygame.Rect.__setattr__(self.abs_hitbox, "topleft", (self.abs_hitbox.left+dx, self.abs_hitbox.top+dy))
                else:
                    self._hitbox = ProtectedHitbox(self.rect.clip(self.window))
                    pygame.Rect.__setattr__(self.abs_hitbox, "topleft", (self.parent.abs.left + self.hitbox.left, self.parent.abs.top + self.hitbox.top))
                    pygame.Rect.__setattr__(self.auto_hitbox, "topleft", (self.hitbox.left - self.rect.left, self.hitbox.top - self.rect.top))
                    if old_hitbox[2:] != self.hitbox.size:
                        pygame.Rect.__setattr__(self.abs_hitbox, "size", self.hitbox.size)
                        pygame.Rect.__setattr__(self.auto_hitbox, "size", self.hitbox.size)
            else:
                pygame.Rect.__setattr__(self.hitbox, "topleft", self.topleft)
                pygame.Rect.__setattr__(self.abs_hitbox, "topleft", self.abs.topleft)
                if debug_with_assert: assert self.auto_hitbox.topleft == (0, 0), self.auto_hitbox

            if debug_with_assert:
                try:
                    if self.window:
                        assert self.hitbox == self.rect.clip(self.window),\
                            "window={}, follow={}, rect={}, hitbox={}" \
                            "".format(self.window, self.window.follow_movements, self.rect, self.hitbox)
                    assert self.abs_hitbox.topleft == (self.parent.abs.left + self.hitbox.left, self.parent.abs.top + self.hitbox.top)
                    assert self.auto_hitbox.topleft == (self.hitbox.left - self.rect.left, self.hitbox.top - self.rect.top)
                except AssertionError as e:
                    raise e

            self.signal.MOTION.emit(dx, dy)

            # We reset the asked_pos after the MOTION in order to allow cycles of origin referecing
            if debug_with_assert: assert not self.has_locked.origin
            self.origin._reset_asked_pos()

            if self.is_visible:
                self.parent._warn_change(self.hitbox.union(old_hitbox))

    def _update_from_parent_movement(self):

        # Note : paint_lock is hold by the parent
        new_pos = self.origin.pos
        old_pos = getattr(self.rect, self.origin.location)
        if self.has_locked.origin:
            self.origin.unlock()
            self._move(dx=new_pos[0]-old_pos[0], dy=new_pos[1]-old_pos[1])
            self.origin.lock()
        else:
            self._move(dx=new_pos[0]-old_pos[0], dy=new_pos[1]-old_pos[1])

    def can_be_resized_at(self, size):

        if self.has_locked.width and size[0] != self.rect.w:
            return False
        if self.has_locked.height and size[1] != self.rect.h:
            return False

        return True

    def clamp_ip(self, Rect):

        raise PermissionError("not implemented")

    def inflate_ip(self, x, y):

        raise PermissionError("not implemented")

    def collidemouse(self):

        return self.abs_hitbox.collidepoint(mouse.pos)

    def lock_height(self, locked=True):
        self.has_locked.height = locked
        self.has_locked.size = self.has_locked.height and self.has_locked.width

    def lock_size(self, locked=True):
        self.has_locked.height = locked
        self.has_locked.width = locked
        self.has_locked.size = locked

    def lock_width(self, locked=True):
        self.has_locked.width = locked
        self.has_locked.size = self.has_locked.height and self.has_locked.width

    def move(self, dx=0, dy=0):

        if dx == 0 and dy == 0: return
        if self.has_locked.origin: return
        self._move(dx, dy)

    def move_at(self, value, key="topleft"):

        accepted = ("x", "y", "centerx", "centery", "top", "bottom", "left", "right", "topleft", "midtop", "topright", "midleft",
                       "center", "midright", "bottomleft", "midbottom", "bottomright")
        assert key in accepted, f"key '{key}' is not a valid rect position (one of {accepted})"

        if self.has_locked.origin:
            return

        old = getattr(self.rect, key)

        if old == value:
            return

        if isinstance(value, (int, float)):

            if key in ("x", "centerx", "left", "right"):
                self._move(value - old, 0)
            else:
                if debug_with_assert: assert key in ("y", "centery", "top", "bottom")
                self._move(0, value - old)

        else:
            self._move(value[0] - old[0], value[1] - old[1])

    def reference(self, pos):
        """
        Require a position relative to the application application
        Return the same position relative to the component's parent topleft
        """
        """if len(pos) == 4:
            rect = pos
            return self.reference((rect[0], rect[1])) + (rect[2], rect[3])"""

        return self.parent.abs_hitbox.referencing(pos)

    def set_window(self, window, follow_movements=None):
        """window is a rect relative to the parent

        follow_movements default to False"""

        if window is not None:
            if window == self.window: return
            assert is_typed_iterable(window, int, 4) or isinstance(window, pygame.Rect),\
                "Wrong recstyle object : {}".format(window)
            self._window = _Window(window)
            if follow_movements is not None:
                self._window.follow_movements = bool(follow_movements)

            old_pos = self.hitbox.topleft
            old_size = self.hitbox.size
            self._hitbox = ProtectedHitbox(self.rect.clip(self.window))
            pygame.Rect.__setattr__(self.abs_hitbox, "topleft", (self.parent.abs.left + self.hitbox.left, self.parent.abs.top + self.hitbox.top))
            # NOTE : should auto_hitbox.topleft be (0, 0) or the difference between self.pos and self.window.topleft ?
            pygame.Rect.__setattr__(self.auto_hitbox, "topleft", (self.hitbox.left - self.rect.left, self.hitbox.top - self.rect.top))
            pygame.Rect.__setattr__(self.abs_hitbox, "size", self.hitbox.size)
            pygame.Rect.__setattr__(self.auto_hitbox, "size", self.hitbox.size)

            if old_pos != self.hitbox.topleft:
                self.signal.MOTION.emit(self.left - old_pos[0], self.top - old_pos[1])
            if old_size != self.size:
                self.signal.RESIZE.emit(old_size)

            if debug_with_assert:
                assert self.abs_hitbox.left == self.parent.abs.left + self.hitbox.left
                assert self.abs_hitbox.top == self.parent.abs.top + self.hitbox.top
        else:
            self._window = None

        self.signal.NEW_WINDOW.emit()
        self.parent._warn_change(self.rect)  # rect is to cover all possibilities

        if debug_with_assert and window is not None:
            assert self.hitbox == self.rect.clip(self.window),\
                "window={}, rect={}, hitbox={}".format(self.window, self.rect, self.hitbox)

    # TODO : rework
    def config_margin(self, margin=None, left=None, top=None, right=None, bottom=None):

        if margin is not None:
            assert left is None
            assert top is None
            assert right is None
            assert bottom is None
            if hasattr(margin, "__iter__"):
                return self.set_margin(left=margin[0], top=margin[1], right=margin[2], bottom=margin[3])
            else:
                return self.set_margin(left=margin, top=margin, right=margin, bottom=margin)

        if left is not None: self.margin.left = left
        if top is not None: self.margin.top = top
        if right is not None: self.margin.right = right
        if bottom is not None: self.margin.bottom = bottom

    def update_pos(self):

        new_pos = self.origin.pos
        old_pos = getattr(self.rect, self.origin.location)
        if new_pos == old_pos: return
        self._move(dx=new_pos[0] - old_pos[0], dy=new_pos[1] - old_pos[1])


class HasProtectedSurface:

    def __init__(self, surface):

        assert isinstance(surface, pygame.Surface)

        """
        surface is the component's image

        Duing the component life, the size of surface can NOT differ from
        the size of the hitbox, it is protected thanks to this classe methods and ProtectedSurface
        
        NEW_SURF is emitted right after set_surface()
        """
        self._surface = surface
        self.create_signal("NEW_SURF")

    surface =            property(lambda self: self._surface)

    def _set_surface(self, surface):

        with paint_lock:
            self._surface = surface
            size = surface.get_size()
            pygame.Rect.__setattr__(self.rect, "size", size)
            pygame.Rect.__setattr__(self.abs_rect, "size", size)
            pygame.Rect.__setattr__(self.auto_rect, "size", size)
            if self.window:
                size = self.rect.clip(self.window).size
            pygame.Rect.__setattr__(self.hitbox, "size", size)
            pygame.Rect.__setattr__(self.abs_hitbox, "size", size)
            pygame.Rect.__setattr__(self.auto_hitbox, "size", size)
            self.update_pos()

    def set_surface(self, surface):

        assert isinstance(surface, pygame.Surface), surface

        if self.has_locked.height and self.rect.height != surface.get_height():
            raise PermissionError("Wrong surface : {} (this component's surface height is locked at {})".format(surface, self.h))

        if self.has_locked.width and self.rect.width != surface.get_width():
            raise PermissionError("Wrong surface : {} (this component's surface width is locked at {})".format(surface, self.w))

        with paint_lock:
            if self.rect.size != surface.get_size():

                old_hitbox = tuple(self.hitbox)
                old_size = self.rect.size
                self._set_surface(surface)
                self.signal.RESIZE.emit(old_size)

                if self.is_visible:
                    self.parent._warn_change(self.hitbox.union(old_hitbox))

            else:

                self._surface = surface
                if self.is_visible:
                    self.parent._warn_change(self.hitbox)

            self.signal.NEW_SURF.emit()

    def subsurface(self, rect):
        """
        If a modification is set on a subsurface, this will change the parent surface,
        but this change won't be visible on screen.
        """

        return self.surface.subsurface(rect).copy()


class Widget(HasStyle, Communicative, HasProtectedSurface, HasProtectedHitbox,
             metaclass=MetaPaintLocker):
    """
    Abstract class for the elements of the screen
    A component can be visible, hidden, static, interactive, dynamic...
    Every component is child of one parent (baopig bio-logic, hmmm...)

    Every Widget have to make its surface at its creation
    """
    STYLE = StyleClass()
    STYLE.create(
        pos = (0, 0),
        pos_location = "topleft",
        pos_ref = None,  # default is parent
        pos_ref_location = "topleft",
        pos_from_ref_hitbox = False,
    )
    STYLE.set_type("pos_location", WidgetLocation)
    STYLE.set_type("pos_ref_location", WidgetLocation)

    def __init__(self, parent, surface, pos=None, origin=None, layer=None, name=None, row=None, col=None, **options):

        if name is None: name = "NoName"
        if isinstance(layer, str): layer = parent.layers_manager.get_layer(layer)

        assert hasattr(parent, "_warn_change")
        assert isinstance(name, str)

        """name is a string who may help to identify components"""
        # defined here so 'name' is first in self.__dict__
        self._name = name

        if "theme" in options:
            theme = options.pop("theme")
            assert isinstance(theme, Theme)
            if not theme.issubtheme(parent.theme):
                raise PermissionError("Must be an parent sub-theme")
        else:
            theme = parent.theme.subtheme()

        HasStyle.__init__(self, theme)

        if "style" in options:
            style = options.pop("style")
            if style:
                self.style.modify(**style)

        Communicative.__init__(self)
        HasProtectedSurface.__init__(self, surface)

        """parent is the Container who contains this Widget"""
        self._parent = parent

        """app is the Application who manages the scenes"""
        self._app = parent.app

        """a weakref will return None after widget.kill()"""
        self._weakref = WeakRef(self)

        """
        A visible component will be rendered on screen
        An invisible component will not be rendered
        
        At appearing, self.signal.SHOW is emitted
        At disappearing, self.signal.HIDE is emitted
        """
        self._is_visible = True
        if "visible" in options:
            self._is_visible = options.pop("visible")
        self.create_signal("SHOW")
        self.create_signal("HIDE")

        """
        dirty is True while the component's surface have not been updated
        
        Particularly usefull when a hidden component got updated
        We wait until it get visible before rendering it
        
        0 : don't need to be updated
        1 : need to be updated once
        2 : need to be updated as much as possible
        """
        self._dirty = 0

        """
        A sleeping component won't be visible and won't update itself until it awakes
        The memory is an object who remember all the changements the component should
        have done while it is sleeping
        When it awakes, it will operate all the changements
        
        Exemple : you can do
        
            comp.move_at((4, 4))
            comp.asleep()
            comp.move_at((10, 10))
            comp.wake()
            
            and the comp will reappear positionned at (10, 10)
        This is very usefull for components who are often hidden and shown
        """
        self._is_sleeping = False
        self._memory = Object(  # TODO : requests_at_awake
            need_appear=None,
            need_start_animation=None,
        )
        self.create_signal("ASLEEP")
        self.create_signal("WAKE")

        """A widget can have this attributes locked"""
        self._has_locked = Object(
            origin=False,
            width=False,
            height=False,
            size=False,
            visibility=False,
        )

        """
        A Widget have a lot of relations with other components, it is hard to remove
        all of them in order to delete the component
        
        Here's why listof_packedfunctions parameter :
        When a component own a method stored in a PackedFunction, if it don't directly
        own the PackedFunction, we need to remember that the component is the method
        owner and in wich PackedFunction that method is stored. This being done, we can
        remove the method from the PackedFunction in order to properly kill the component
        """
        self.create_signal("KILL")

        """
        The scene is always the same
        Since self is a scene's child, it doesn't need to only keep a weakref to it
        """
        self.__scene = parent.scene

        if pos:
            self.style.modify(pos=pos)
        self._col = None
        self._row = None
        if (col is not None) or (row is not None):
            """col and row attributes are destinated for GridLayer"""
            if (pos is not None) or \
                    ("pos_location" in options) or \
                    ("pos_ref" in options) or \
                    ("pos_ref_location" in options):
                raise PermissionError("When the layer manages the widget's position, "
                                      "all you can define is row, col and sticky")
            if col is None: col = 0
            if row is None: row = 0
            assert isinstance(col, int) and col >= 0
            assert isinstance(row, int) and row >= 0
            self._col = col
            self._row = row
        self._sticky = None
        if "sticky" in options:
            self._sticky = options.pop("sticky")
            self.style.modify(
                pos_location = self._sticky,
                pos_ref_location = self._sticky
            )
        if "pos_location" in options:
            self.style.modify(pos_location = options.pop("pos_location"))
        if "pos_ref" in options:
            self.style.modify(pos_ref = options.pop("pos_ref"))
        if "pos_ref_location" in options:
            self.style.modify(pos_ref_location = options.pop("pos_ref_location"))

        # TODO : center=(34, 45) instead of pos=(34, 45), pos_location="center"

        """
        A component is stored inside its parent, at one layer. A layer is identified
        via a string name. Default layer is defined by the parent. You can change a
        component layer via container.layer(component, layer_name) where layer_name
        references an existing layer.
        """
        layer_level = None
        if "layer_level" in options:
            assert layer is None, "Cannot define a layer AND a layer_level"
            layer_level = options.pop("layer_level")
        if layer is None:
            layer = parent.layers_manager.find_layer_for(self, layer_level)
        self._layer = layer
        # if col is not None:  # or col and row are at None, or they both are defined
        #     self.layer.

        # INITIALIZATIONS
        HasProtectedHitbox.__init__(self)
        self.parent._add_child(self)

        if debug_new:
            print("NEW COMPONENT :", self)

        if "touchable" in options:
            if options["touchable"] is False:
                self.set_nontouchable()
            options.pop("touchable")

        if options:
            raise ValueError(f"Unused options : {options}")

        def paint_decorator(widget, paint):
            functools.wraps(paint)
            def wrapped_func(*args, **kwargs):
                res = paint(*args, **kwargs)
                widget.parent._warn_change(widget.hitbox)
                return res
            return wrapped_func
        self.paint = paint_decorator(self, self.paint)

    def __del__(self):

        if debug_del:
            print("DELETE :", self)
            global _is_alive
            _is_alive = False

    def __repr__(self):
        """
        Called by repr(self)

        :return: str
        """
        # This complicated string creation is avoiding the o.__repr__()
        # in order to avoid representation of components
        return "<{}(name='{}', parent='{}', hitbox={})>".format(
            self.__class__.__name__, self.name, self.parent.name, self.hitbox
        )

    def __str__(self):
        """
        Called by str(self)
        """
        return f"{self.__class__.__name__}(name={self.name})"

    # GETTERS ON PROTECTED FIELDS
    app = application = property(lambda self: self._app)
    col =               property(lambda self: self._col)
    dirty =             property(lambda self: self._dirty)
    has_locked =        property(lambda self: self._has_locked)
    is_alive =          property(lambda self: self._weakref._comp is not None)
    is_awake =          property(lambda self: not self._is_sleeping)
    is_dead =           property(lambda self: self._weakref._comp is None)
    is_hidden =         property(lambda self: not self._is_visible)
    is_sleeping =       property(lambda self: self._is_sleeping)
    is_visible =        property(lambda self: self._is_visible)
    layer =             property(lambda self: self._layer)
    name =              property(lambda self: self._name)
    parent =            property(lambda self: self._parent)
    row =               property(lambda self: self._row)
    scene =             property(lambda self: self.__scene)
    sticky =            property(lambda self: self._sticky)

    def asleep(self):

        self.parent.asleep_child(self)

    def copy(self):
        """
        Abstract method
        """
        raise PermissionError("The copy() method has not been overriden")

    def depend_on(self, comp):

        self.connect("kill", comp.signal.KILL)

    def get_weakref(self, callback=None):
        """
        A weakref is a reference to an object
        callback is a function called when the component die

        Example :
            weak_ref = my_comp.get_weakref()
            my_comp2 = weak_ref()
            my_comp is my_comp2 -> return True
            del my_comp
            print(my_comp2) -> return None
        """
        if callback is not None:
            assert callable(callback)
            self.signal.KILL.connect(callback)
        return self._weakref

    def hide(self):
        """
        # NOTE : if overriden, be carefull -> you should call super().hide()
        """

        if self.has_locked.visibility: return

        if not self.is_visible:
            return

        if self.is_sleeping:
            self._memory.need_appear = False
            return

        self._is_visible = False

        if self == mouse.linked_comp:
            mouse._unlink()
        elif self.collidemouse():
            mouse.update_hovered_comp()

        self.parent._warn_change(self.hitbox)
        self.signal.HIDE.emit()

    def kill(self):

        if not self.is_alive: return
        with paint_lock:
            try:
                if debug_with_assert:
                    if self.is_awake:
                        assert self in self.parent.children
                    else:
                        assert self in self.parent.children.sleeping
                self.parent._remove_child(self)
                self.disconnect()
                self.signal.KILL.emit(self._weakref)
                self._weakref._comp = None
                if debug_with_assert: assert self not in self.parent.children
            except Exception as e:
                raise e

        if debug_kill:
            print("KILL :", self)

        if debug_del:
            global _is_alive
            _is_alive = True
            # import sys
            # warn_msg = "This component is not entirely deleted : {} (dependencies : {})"\
            #                    "".format(self, sys.getrefcount(self))
        del self
        if debug_del:
            if _is_alive:
                gc.collect()
                # if _is_alive:
                #     LOGGER.warning(warn_msg)
            _is_alive = True

    def lock_visibility(self, locked=True):

        if self.has_locked.visibility == locked: return
        self.has_locked.visibility = locked

    def move_behind(self, comp):

        if debug_with_assert: assert comp is not self
        self.layer.move_comp1_behind_comp2(comp1=self, comp2=comp)

    def move_in_front_of(self, comp):

        if debug_with_assert: assert comp is not self
        self.layer.move_comp1_in_front_of_comp2(comp1=self, comp2=comp)

    def paint(self):
        """
        This method is made for being overriden

        In your implementation, you can update the component's surface.
        If you use send_paint_request(), this method will be called when the next frame is rendered
        In fact, you can use paint() at any moment, but it is more efficient
        to update it through send_paint_request() if it is changing very often
        """
        pass

    def send_paint_request(self):

        if self._dirty == 0:
            self.parent._dirty_child(self, 1)

    def set_nontouchable(self):
        """Cannot go back"""

        # TODO : documentation
        # TODO : rework
        need_mouse_update = self.collidemouse()
        self.collidemouse = lambda: False
        if need_mouse_update: mouse.update_hovered_comp()

    def set_dirty(self, dirty):
        """
        Works as pygame.DirtySprite :

        if set to 1, it is repainted and then set to 0 again
        if set to 2 then it is always dirty (repainted each scene, flag is not reset)
        0 means that it is not dirty and therefor not repainted again
        """

        if debug_with_assert: assert dirty in (0, 1, 2)
        self.parent._dirty_child(self, dirty)

    def set_visibility(self, visible):

        if visible: self.show()
        else:       self.hide()

    def show(self):
        """
        If the component was awake and hidden, it will appear on the screen
        """

        if self.has_locked.visibility: return

        if self.is_visible:
            return

        if self.is_sleeping:
            self._memory.need_appear = True
            return

        self._is_visible = True

        if self.collidemouse():
            mouse.update_hovered_comp()
        self.parent._warn_change(self.hitbox)
        if self._dirty:
            self.send_paint_request()
        self.signal.SHOW.emit()

    def swap_layer(self, layer):

        if isinstance(layer, str): layer = self.parent.layers_manager.get_layer(layer)
        self.parent.layers_manager.swap_layer(self, layer)

    def toggle_sleeping(self):

        if self.is_sleeping:
            self.wake()
        else:
            self.asleep()

    def toggle_visibility(self):

        if self.is_visible:
            self.hide()
        else:
            self.show()

    @staticmethod
    def ivisibles_from(list):
        for comp in list:
            if comp.is_visible:
                yield comp

    def wake(self):

        self.parent.wake_child(self)





from baopig import *


class UT_Focusable_Frame(Zone):
    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)

        class PosButton(Button):
            def __init__(self, parent, pos):
                Button.__init__(self, parent, pos=pos, text="({}, {})".format(*pos))

        b1 = PosButton(self, pos=(10, 10))
        b0 = PosButton(self, pos=(10, 45))
        b0 = PosButton(self, pos=(80, 10))
        b2 = PosButton(self, pos=(80, 45))
        b0 = PosButton(self, pos=(10, 80))
        b0 = PosButton(self, pos=(80, 80))
        Dragable.set_dragable(b0)

        z = Zone(self, pos=(-5, 5), sticky="topright",
                 size=(self.w-155, self.h-10), background_color=(10, 30, 20, 128))
        import random
        for i in range(10):
            x = random.randrange(z.w - b0.w)
            y = random.randrange(z.h - b0.h)
            PosButton(z, pos=(x, y))

ut_scenes = [
    UT_Focusable_Frame,
]


if __name__ == "__main__":
    from baopig.unit_tests.TesterScene import TesterScene
    app = Application()
    for scene in ut_scenes:
        TesterScene(app, scene)
    app.launch()



from baopig._lib import Scrollable, ScrollableByMouse, Size, MarginType
from baopig._lib import Zone, Rectangle, LayersManager
from .slider import Slider, SliderBar, SliderBloc


class ScrollSlider(Slider):
    STYLE = Slider.STYLE.substyle()
    STYLE.modify(
        width=10,
        height=0,
    )

    def __init__(self, scroller, axis, style):

        assert isinstance(scroller, Scrollable)

        self.inherit_style(scroller, style)
        assert self.style["height"] == 0, \
            "You cannot define height on a ScrollSlider, it is automatically generated"

        border = MarginType(self.theme[SliderBar]["border_width"])
        scroller_size = scroller.window[2:]
        width = self.style["width"]
        self._parent = scroller  # needed in get_length()
        self._axis = axis  # needed in get_length()
        length = self.get_length()
        if axis == "x":
            bloc_style = {"width": length, "height": width}
            pos = (0, border.bottom)
            sticky = "bottom"
            size = (scroller_size[0] + border.left + border.right, width)
        else:
            bloc_style = {"width": width, "height": length}
            pos = (border.right, 0)
            sticky = "right"
            size = (width, scroller_size[1] + border.top + border.bottom)
        bloc_style["border_width"] = 0
        self.set_style_for(SliderBloc, **bloc_style)

        Slider.__init__(
            self, scroller, 0, 100, step=1e-9, axis=axis,
            pos=pos, sticky=sticky, bar_size=size, pos_from_ref_hitbox=True,
            layer_level=LayersManager.FOREGROUND
        )

    def get_length(self):

        if self.axis == "y":
            return self.parent.hitbox.height / self.parent.height * self.parent.window[3]
        else:
            return self.parent.hitbox.width / self.parent.width * self.parent.window[2]


class ScrollableZone(Zone, ScrollableByMouse):
    STYLE = Zone.STYLE.substyle()
    STYLE.create(
        scrollslider_class=ScrollSlider
    )
    STYLE.set_constraint("scrollslider_class", lambda val: issubclass(val, ScrollSlider))

    def __init__(self, parent, window_size, bar_style=None, **options):

        Zone.__init__(self, parent, **options)

        self._scrollaxis = ""
        self.scrollsliders = []
        scrollslider_class = self.style["scrollslider_class"]

        self.set_window(self.pos + Size(window_size))
        assert self.rect.contains(self.rect.__class__(self.window))  # TODO : remove
        if self.width > window_size[0]:
            self._scrollaxis += "x"
            self.scrollsliders.append(scrollslider_class(self, "x", bar_style))
        if self.height > window_size[1]:
            self._scrollaxis += "y"
            self.scrollsliders.append(scrollslider_class(self, "y", bar_style))
        ScrollableByMouse.__init__(self, self.scrollaxis)

    def get_scroll_pourcent(self, axis):

        assert axis in "x", "y"

        if axis == "x":
            return self.window[0] / (self.width - self.window[2])
        else:
            return self.window[1] / (self.height - self.window[3])





import pygame
from baopig._debug import debug_screen_updates, debug_with_assert
from baopig.io.logging import LOGGER
from baopig._lib import Box, Layer, paint_lock, Widget, Size


class Zone(Box):  # TODO : BoxModel

    STYLE = Box.STYLE.substyle()
    STYLE.modify(
        width = 100,
        height = 100,
    )

    def __init__(self, parent, **options):

        Box.__init__(self, parent, **options)

    def divide(self, side, width):
        # TODO : rework this function
        if side == "left":
            self.rect.left = width
            self.rect.w -= width
            if self.w <= 0:
                raise ValueError("the new zone shouldn't completly override the central zone")
            zone = Zone((width, self.rect.h))
            comps_to_move = []  # On ne peut iterer sur une liste tout en la modifiant
            for comp in self._components:
                if zone.hitbox.collidepoint(comp.topleft):
                    comps_to_move.append(comp)
                else:
                    comp.left -= width
            for comp in comps_to_move:
                zone.append(comp)
                self.remove(comp)
        return zone


class SubZone(Zone):  # TODO : SubScene ? with rects_to_update ?
    # TODO : solve : when update on a SubZone whose parent is a scene, the display isn't updated
    """A SuzZone is an optimized Zone, its surface is a subsurface of its parent (cannot have transparency)"""

    def __init__(self, parent, **kwargs):

        Zone.__init__(self, parent, **kwargs)
        try:
            self._surface = self.parent.surface.subsurface(self.pos + self.size)
        except ValueError:
            assert not self.parent.auto.contains(self.rect)
            raise PermissionError("A SubZone must fit inside its parent")
        if debug_with_assert: assert self.surface.get_parent() is self.parent.surface

        self.connect("_update_subsurface", self.parent.signal.NEW_SURF)
        self.connect("_update_subsurface", self.signal.MOTION)

    def _update_subsurface(self):

        with paint_lock:
            try:
                Widget.set_surface(self, self.parent.surface.subsurface(self.pos + self.size))
            except ValueError:
                assert not self.parent.auto.contains(self.rect)
                Widget.set_surface(self, self.parent.surface.subsurface(
                    pygame.Rect(self.rect).clip(self.parent.auto)))  # resize the subzone
            if debug_with_assert: assert self.surface.get_parent() is self.parent.surface

    def _flip(self):
        """Update all the surface"""

        if self.is_hidden:  return

        with paint_lock:

            self._flip_without_update()

            # optimization
            if self.parent is self.scene:
                pygame.display.update(self.hitbox)
            else:
                self.parent.parent._warn_change(
                    (self.parent.left + self.hitbox.left, self.parent.top + self.hitbox.top) + tuple(self.hitbox.size)
                )
            if debug_with_assert: assert self.surface.get_parent() is self.parent.surface

    def _warn_parent(self, rect):
        """Request updates at rects referenced by self"""

        rect = (self.left + rect[0], self.top + rect[1]) + tuple(rect[2:])

        # because of subsurface, we can skip self.parent._update_rect()
        if self.parent is self.scene:
            pygame.display.update(rect)
        else:
            self.parent._warn_parent(rect)
        if debug_with_assert: assert self.surface.get_parent() is self.parent.surface

    def resize(self, w, h):

        if self.has_locked.width: w = self.w
        if self.has_locked.height: h = self.h
        if (w, h) == self.size: return

        with paint_lock:
            try:
                Widget.set_surface(self, self.parent.surface.subsurface(self.pos + (w, h)))
            except ValueError:
                assert not self.parent.auto.contains(self.rect)
                raise PermissionError("A SubZone must fit inside its parent")
                Widget.set_surface(self, self.parent.surface.subsurface(
                    pygame.Rect(self.pos + (w, h)).clip(self.parent.auto)))
            self._flip_without_update()

        # print("RESIZED", self, "at", size)





import pygame
from pygame.image import *  # so baopig.image contains pygame.image
from .resizable import ResizableWidget


class Image(ResizableWidget):

    # TODO : self.tiled instead of parameter in resize()

    def __init__(self, parent, image, pos=None, w=None, h=None, **kwargs):
        """
        Cree une image

        If w or h parameters are filled, width or height of image argument
        are respectively resized
        """

        assert isinstance(image, pygame.Surface), "image must be a Surface"

        image_size = image.get_size()
        if w is None: w = image_size[0]
        if h is None: h = image_size[1]
        if image_size != (w, h):
            surface = pygame.transform.scale(image, (w, h))
        else:
            surface = image.copy()

        ResizableWidget.__init__(
            self,
            parent=parent,
            surface=surface,
            pos=pos,
            **kwargs
        )

    def collidemouse_alpha(self):  # TODO
        pass

    def resize(self, w, h, tiled=False):

        if tiled:

            surface = pygame.Surface((w, h), pygame.SRCALPHA)
            surface.blit(self.surface, (0, 0))

            if w > self.w:
                for i in range(int(w / self.w)):
                    surface.blit(self.surface, (self.w * (i + 1), 0))

            if h > self.h:
                row = surface.subsurface((0, 0, w, self.h)).copy()
                for i in range(int(h / self.h)):
                    surface.blit(row, (0, self.h * (i + 1)))

            self.set_surface(surface)

        else:
            self.set_surface(pygame.transform.scale(self.surface, (w, h)))




from baopig import *


# ------ TESTS TO ACHIEVE ------

# --- SECTION 1 : Selector ---
# TEST 01 : A Selector can be focused
# TEST 02 : When receive Ctrl + A, calls the 'select_all' method
# TEST 03 : The default implementation of 'select_all' selects all selectable children
# TEST 04 : When receive link_motion, it creates a SelectionRect
# TEST 05 : When receive Ctrl + C, calls the 'copy' method
# TEST 06 : The default implementation of 'copy' collects data from selected Selectables, join into a string and send it to the clipboard
# TEST 07 : When receive Ctrl + X, calls the 'cut' method (no default implementation)
# TEST 08 : When receive Ctrl + V, calls the 'paste' method (no default implementation)
# TEST 09 : The call of 'disable_selecting' will deactivate the selection ability

# ------ SECTION 2 : Selectable ------
# TEST 01 : A Selectable's selector is the youngest Selector in Selectable's family tree
# TEST 02 : For each selection rect movement, if it collide with a selectable, it calls the 'select' method
# TEST 03 : At the time selection rect don't collide with a selectable anymore, 'unselect' is called

# ----- SECTION 3 : The selection rectangle -----
# TEST 01 : A drag on a focused Selector create a selection rect
# TEST 02 : A released drag will hide the selection rect
# TEST 03 : The selection rect can be configured throught passing a subclass of SelectionRect in the Selector constructor
# TEST 04 : The visibility of the selection rect can be edited -> set_selectionrect_visibility
# TEST 05 : When setting an end position for the selection rect, a temporary visibility can be given in argument
# TEST 06 : The selection rect is always fitting inside its parent


class UT_Selectable(Rectangle, Selectable):

    def __init__(self, parent, pos):

        Rectangle.__init__(self, parent, color="blue", size=(30, 30), pos=pos)
        Selectable.__init__(self)
        self.hightlighter = None
        self.timer = None

    def select(self):

        self.set_color("blue4")
        self.highlight("green")

    def unselect(self):

        self.set_color("blue")
        self.highlight("red")

    def highlight(self, color):

        def timeout():
            self.hightlighter.kill()
            self.hightlighter = None

        if self.hightlighter is not None:
            if self.hightlighter.color == Color(color):
                return
            self.timer.cancel()
            timeout()
        self.hightlighter = Highlighter(self.parent, self, color, 2)
        self.timer = Timer(.4, timeout)
        self.timer.start()


class UT_Selector(Zone, Selector):

    def __init__(self, *args, **kwargs):

        Zone.__init__(self, *args, **kwargs)
        Selector.__init__(self)
        self.enable_selecting(True)


class UT_Selections_Frame(UT_Selector):

    def __init__(self, *args, **kwargs):
        UT_Selector.__init__(self, *args, **kwargs)

        z = UT_Selector(self, size=(self.w/3, self.h - 20), background_color="gray",
                        pos=("50%", 10), pos_location="midtop")
        UT_Selectable(z, (10, 10))
        UT_Selectable(z, (50, 10))
        UT_Selectable(z, (90, 10))
        Text(z, "I am selectable", pos=(10, 50))
        TextEdit(z, pos=(10, 75), width=z.w-20)  # TODO : Scrollable

        z2 = UT_Selector(z, size=(z.w-20, (z.h-40)/3), background_color=(128, 128, 128),
                         pos=(10, "50%"), pos_location="midleft")
        z2.enable_selecting(False)
        UT_Selectable(z2, (10, 10))
        UT_Selectable(z2, (50, 10))
        UT_Selectable(z2, (90, 10))
        Text(z2, "I am not selectable", pos=(10, 50))
        TextEdit(z2, pos=(10, 75), width=z2.w-20)

        z3 = UT_Selector(z, size=(z.w-20, (z.h-40)/3), background_color=(128, 128, 128, 200),
                         pos=(10, -10), pos_location="bottom", pos_ref_location="bottom")
        z3.set_selectionrect_visibility(False)
        UT_Selectable(z3, (10, 10))
        UT_Selectable(z3, (50, 10))
        UT_Selectable(z3, (90, 10))
        Text(z3, "Selection rectangle ?", pos=(10, 50))
        TextEdit(z3, pos=(10, 75), width=z3.w-20)

        UT_Selectable(self, (10, 10))
        UT_Selectable(self, (50, 10))
        UT_Selectable(self, (90, 10))
        Text(self, "I am selectable", pos=(10, 50))
        TextEdit(self, pos=(10, 75), width=z.w-20)

        UT_Selectable(self, (z.right + 10, 10))
        UT_Selectable(self, (z.right + 50, 10))
        UT_Selectable(self, (z.right + 90, 10))
        Text(self, "I am selectable", pos=(z.right + 10, 50))
        TextEdit(self, pos=(z.right + 10, 75), width=z.w-20)

        self.load_sections()

    def load_sections(self):

        self.parent.add_section(
            title="Selector",
            tests=[
                "A Selector can be focused",
                "When receive Ctrl + A, calls the 'select_all' method",
                "The default implementation of 'select_all' selects all selectable children",
                "When receive link_motion, it creates a SelectionRect",
                "When receive Ctrl + C, calls the 'copy' method",
                "The default implementation of 'copy' collects data from selected Selectables, join into a string and send it to the clipboard",
                "When receive Ctrl + X, calls the 'cut' method (no default implementation)",
                "When receive Ctrl + V, calls the 'paste' method (no default implementation)",
                "The call of 'disable_selecting' will deactivate the selection ability",
            ]
            # TODO : zone=Zone(...)
        )

        self.parent.add_section(
            title="Selectable",
            tests=[
                "A Selectable's selector is the youngest Selector in Selectable's family tree",
                "For each selection rect movement, if it collide with a selectable, it calls the 'select' method",
                "At the time selection rect don't collide with a selectable anymore, 'unselect' is called",
            ]
        )

        self.parent.add_section(
            title="The selection rectangle",
            tests=[
                "A drag on a focused Selector create a selection rect",
                "A released drag will hide the selection rect",
                "The selection rect can be configured throught passing a subclass of SelectionRect in the Selector constructor",
                "The visibility of the selection rect can be edited -> set_selectionrect_visibility",
                "When setting an end position for the selection rect, a temporary visibility can be given in argument",
                "The selection rect is always fitting inside its parent",
            ]
        )


ut_scenes = [
    UT_Selections_Frame,
]


if __name__ == "__main__":
    from baopig.unit_tests.TesterScene import TesterScene
    app = Application()
    for scene in ut_scenes:
        TesterScene(app, scene)
    app.launch()




import time
import pygame
from baopig.pybao.objectutilities import Object, History, debug
from baopig.communicative import Communicative
from .logging import LOGGER


class MouseEvent(Object):

    def __init__(self, signal, **kwargs):

        Object.__init__(self, signal=signal, type=getattr(mouse, signal), **kwargs)
        mouse.last_event = self

    def __str__(self):
        global __mouse_signals
        return "<MouseEvent({}-{} {})>".format(
            self.type,
            list(mouse._signals.keys())[list(mouse._signals.values()).index(self.type)],
            self.__dict__
        )
    __repr__ = __str__


iterator = (i for i in range(60, 100))
class _Mouse(Communicative):

    LEFTBUTTON_DOWN = next(iterator)
    WHEELBUTTON_DOWN = next(iterator)
    RIGHTBUTTON_DOWN = next(iterator)

    LEFTBUTTON_UP = next(iterator)
    WHEELBUTTON_UP = next(iterator)
    RIGHTBUTTON_UP = next(iterator)

    SCROLL = next(iterator)

    MOTION = next(iterator)
    DRAG = next(iterator)
    RELEASEDRAG = next(iterator)

    _signals = {

        "LEFTBUTTON_DOWN": LEFTBUTTON_DOWN,
        "WHEELBUTTON_DOWN": WHEELBUTTON_DOWN,
        "RIGHTBUTTON_DOWN": RIGHTBUTTON_DOWN,

        "LEFTBUTTON_UP": LEFTBUTTON_UP,
        "WHEELBUTTON_UP": WHEELBUTTON_UP,
        "RIGHTBUTTON_UP": RIGHTBUTTON_UP,

        "SCROLL": SCROLL,

        "MOTION": MOTION,
        "DRAG": DRAG,
        "RELEASEDRAG": RELEASEDRAG,
    }

    def __init__(self):

        Communicative.__init__(self)

        for signal in self._signals:
            self.create_signal(signal)

        self._pos = (0, 0)
        self._rel = (0, 0)  # Le dernier deplacement de la souris

        # L'etat des bouttons de la souris (il y en a 5)
        # self.button[3] -> is the button 3 pressed ?
        # There is no button 0, so self.button[0] = None
        # For pygame, the 4 and 5 buttons are wheel up and down,
        # so these buttons are implemented as mouse.SCROLL

        self._pressed_button = None  # TODO : why not a tuple of pressed buttons ?

        # [None, 0, 0, 0]  # WARNING : A mouse can have additionnals buttons

        # Using an empty clic avoid testing if the mouse has clicks in memory or not
        empty_clic = Object(
            time=-1,  # heure du clic - en secondes
            button=None,  # numero du bouton
            pos=None)  # position du clic
        historic_size = 3  # Pour le triple-clic
        self.clic_history = History(object, maxlen=historic_size, seq=[empty_clic] * historic_size)

        """
        When the mouse is hovering a Text inside a Button inside a Zone inside a Scene,
        then the Text is hovered
        """
        self._hovered_comp = None

        """
        When the mouse click on a Text inside a Button inside a Zone inside a Scene,
        then the Text is linked
        """
        self._linked_comp = None

        """
        When a drag occurs, the youngest clicked container is selecting
        The selection is a rect from drag_origin to mouse.pos
        """
        self.drag_origin = None

        """
        Permet de savoir si l'utilisateur vient de faire un double-click

        Pour faire un double-clic, il faut :
            - qu'il y ait eu 2 clics en moins de 5 dixiemes de secondes
            - que les 2 clics aient ete fait par le bouton gauche
            - que la souris n'ait pas bougee entre les clics

        has_double_clicked garde la valeur True jusqu'au clic suivant
        """
        self.has_double_clicked = False

        """
        Permet de savoir si l'utilisateur vient de faire un triple-clic

        Pour faire un triple-clic, il faut :
            - qu'il y ait eu 3 clics en moins de 10 dixiemes de secondes
            - que les 3 clics aient ete fait par le bouton gauche
            - que la souris n'ait pas bougee entre les clics

        has_triple_clicked garde la valeur True jusqu'au clic suivant
        """
        self.has_triple_clicked = False

        self._application = None
        self._display = None
        self._is_hovering_display = False
        self.last_event = None

    def __repr__(self):
        return "<Mouse(" + str(self.__dict__) + ")>"

    def __str__(self):
        return "<Mouse('pos': {}, 'pressed_button': {}, 'last_event': {})>" \
               "".format(self.pos, self.pressed_button, self.last_event)

    pos = property(lambda self: self._pos)
    x = property(lambda self: self._pos[0])  # TODO : allow moving mouse by doing mouse.move(dx, dy)
    y = property(lambda self: self._pos[1])  #         -> (for joystick control over mouse)

    application = property(lambda self: self._application)
    scene = property(lambda self: self._application._focused_scene)
    is_hovering_display = property(lambda self: self._is_hovering_display)
    linked_comp = property(lambda self: self._linked_comp)
    hovered_comp = property(lambda self: self._hovered_comp)
    pressed_button = property(lambda self: self._pressed_button)

    def _link(self, comp):

        assert self.is_hovering_display
        assert self.linked_comp is None

        if comp is not None:
            # assert comp.is_visible
            assert not comp.is_linked

            self._linked_comp = comp
            comp.is_linked = True
            self.linked_comp.signal.LINK.emit()

    def _get_pointed_comp(self):

        def get_pointed_comp(cont):

            if cont.collidemouse():
                for layer in reversed(tuple(cont.layers_manager.touchable_layers)):
                    assert layer.touchable
                    for comp in reversed(layer):
                        if comp.is_visible and comp.collidemouse():
                            if hasattr(comp, "children"):
                                return get_pointed_comp(comp)
                            return comp
                return cont

        return get_pointed_comp(self.scene)

    pointed_comp = property(_get_pointed_comp)

    def _hover_display(self):

        if self.is_hovering_display:
            return
        self._is_hovering_display = True
        self.update_hovered_comp()
        # print("HOVER APP")

    def _release_all(self):
        """
        Only called by Application.freeze()
        Release the pressed button
        """
        if self.pressed_button is not None:
            self.receive(Object(type=pygame.MOUSEBUTTONUP, button=self.pressed_button))

    def _unhover_display(self):

        if not self.is_hovering_display:
            return
        self._hover(None)
        self._is_hovering_display = False
        # print("UNHOVER APP")

    def _hover(self, comp):

        # if not self.is_hovering_display:
        #     assert self.hovered_comp is None
        #     return

        if comp is self.hovered_comp: return

        # UNHOVER
        old_hovered = self.hovered_comp
        if self.hovered_comp is not None:
            try:
                assert old_hovered.is_hovered
                old_hovered._is_hovered = False
            except:
                print("Hovering error")

        self._hovered_comp = comp

        # HOVER
        if comp is not None:
            assert comp.is_visible, repr(comp)
            assert not comp.is_hovered
            comp._is_hovered = True

        # SIGNALS
        if old_hovered is not None and hasattr(old_hovered.signal, "UNHOVER"):
            old_hovered.signal.UNHOVER.emit()
        if comp is not None and hasattr(comp.signal, "HOVER"):
            comp.signal.HOVER.emit()

    def _unlink(self):
        """
        This method unlink a Linkable component from the mouse

        It can exceptionnaly be called when a clicked component disappears
        Then the component calls itself this function, trougth Linkable.unlink()
        """

        try:
            assert self.linked_comp.is_linked or self.linked_comp.is_dead
        except AssertionError as e:
            raise e

        comp = self.linked_comp
        self._linked_comp = None

        if comp.is_alive:
            comp.is_linked = False
            comp.signal.UNLINK.emit()
        # While the mouse left button was press, we didn't update hovered_comp
        self.update_hovered_comp()

    def defocus_TBR(self, comp):
        """
        Permit to a component to defocus itself
        """
        assert self.scene.focused_comp == comp
        self.scene._focus(None)

    def get_pos_relative_to(self, comp):

        return comp.abs_rect.referencing(self.pos)

    def is_pressed(self, button):

        return button == self.pressed_button

    def receive(self, event):

        if event.type not in (pygame.MOUSEBUTTONUP, pygame.MOUSEBUTTONDOWN, pygame.MOUSEMOTION):
            LOGGER.warning("Unknown event : {}".format(event))
            return

        if event.type in (pygame.MOUSEBUTTONUP, pygame.MOUSEBUTTONDOWN):

            if event.button not in (1, 2, 3, 4, 5):
                LOGGER.warning("Unknown button id : {} (event : {})".format(event.button, event))
                return

            if self.pressed_button is not None and self.pressed_button != event.button:
                # Another button is already pressed, we skip
                LOGGER.info("Another button is already pressed, we skip")
                return

            if event.type == pygame.MOUSEBUTTONUP:
                if event.button in (4, 5):
                    # It's a wheel end of scrolling, wich is useless
                    return

                if not self.is_pressed(event.button):
                    # This button's DOWN event have been skiped, so we skip it's UP event
                    return

        elif event.type == pygame.MOUSEMOTION:

            if event.rel == (0, 0):
                # Don't care
                return

        # Real work

        if event.type == pygame.MOUSEBUTTONDOWN:

            # ACTUALIZING MOUSE STATE

            if event.button not in (4, 5):  # Ignore wheel
                self.clic_history.append(Object(time=time.time(), button=event.button, pos=event.pos))
                self._pressed_button = event.button

                self.has_double_clicked = \
                    self.clic_history[-2].time > self.clic_history[-1].time - .5 and \
                    self.clic_history[-2].button == self.clic_history[-1].button == 1 and \
                    self.clic_history[-2].pos == self.clic_history[-1].pos

                self.has_triple_clicked = \
                    self.has_double_clicked and \
                    self.clic_history[-3].time > self.clic_history[-1].time - 1 and \
                    self.clic_history[-3].button == self.clic_history[-1].button == 1 and \
                    self.clic_history[-3].pos == self.clic_history[-1].pos

            if event.button == 1:
                # UPDATING CLICKS, FOCUSES, HOVERS...

                def first_focusable_in_family_tree(widget):
                    # The recursivity always come to an end because a Scene is focusable
                    if hasattr(widget, "is_focused") and widget.is_enabled:
                        return widget
                    if widget.parent is widget:
                        return None
                    return first_focusable_in_family_tree(widget.parent)

                # Le focus passe avant le link parce que Linkable est une sous-class de Focusable
                focused = first_focusable_in_family_tree(self._get_pointed_comp())
                self.scene._focus(focused)
                if hasattr(focused, "is_linked"):
                    self._link(focused)
                else:
                    self._link(None)

            # MOUSE EVENTS TRANSMISSION

            if event.button == 1:
                MouseEvent(
                    signal="LEFTBUTTON_DOWN",
                    pos=self.pos,
                )
            elif event.button == 2:
                MouseEvent(
                    signal="WHEELBUTTON_DOWN",
                    pos=self.pos,
                )
            elif event.button == 3:
                MouseEvent(
                    signal="RIGHTBUTTON_DOWN",
                    pos=self.pos,
                )
            elif event.button == 4:
                MouseEvent(
                    signal="SCROLL",
                    direction=1,
                )
                self.update_hovered_comp()
            elif event.button == 5:
                MouseEvent(
                    signal="SCROLL",
                    direction=-1,
                )
                self.update_hovered_comp()

        elif event.type == pygame.MOUSEBUTTONUP:

            # We don't consider an event of a pressed button when another button is already pressed
            # So this is the prevention
            assert self.is_pressed(event.button)

            # ACTUALIZING MOUSE STATE
            self._pressed_button = None

            # MOUSE EVENTS TRANSMISSION
            if event.button == 1:  # release left button

                MouseEvent(
                    signal="LEFTBUTTON_UP",
                    pos=self.pos
                )

            elif event.button == 2:  # release wheel button

                MouseEvent(
                    signal="WHEELBUTTON_UP",
                    pos=self.pos,
                )

            elif event.button == 3:  # release right button

                MouseEvent(
                    signal="RIGHTBUTTON_UP",
                    pos=self.pos,
                )

            if self.drag_origin is not None:
                MouseEvent(
                    signal="RELEASEDRAG",
                    pos=self.pos,
                    origin=self.drag_origin,
                    total_rel=(self.pos[0] - self.drag_origin[0], self.pos[1] - self.drag_origin[1]),
                    button=event.button,
                )
                self.drag_origin = None

            # UPDATING CLICKS, FOCUSES, HOVERS...

            if self.linked_comp:
                self._unlink()

        elif event.type == pygame.MOUSEMOTION:

            # ACTUALIZING MOUSE STATE
            self._pos = event.pos
            if self.drag_origin is None:
                # NOTE : Sometimes, pygame.MOUSEMOTION doesn't have attribute 'buttons'
                if self.pressed_button is not None:
                    self.drag_origin = (self.pos[0] - event.rel[0], self.pos[1] - event.rel[1])

            # MOUSE EVENTS TRANSMISSION
            if self.drag_origin is None:
                self.update_hovered_comp()
                MouseEvent(
                    signal="MOTION",
                    pos=self.pos,
                    rel=event.rel,
                )
            else:
                MouseEvent(
                    signal="DRAG",
                    pos=self.pos,
                    origin=self.drag_origin,
                    rel=event.rel,
                    total_rel=(self.pos[0] - self.drag_origin[0], self.pos[1] - self.drag_origin[1]),
                    button=self.pressed_button
                )
                # HANDLINGS
                if self.pressed_button == 1 and self.linked_comp:  # left button drag
                    self.linked_comp.signal.LINK_MOTION.emit(self.last_event)

        getattr(self.signal, self.last_event.signal).emit(self.last_event)

    def update_hovered_comp(self):

        if self.linked_comp is not None: return
        def first_hoverable_in_family_tree(widget):
            if widget.scene is widget:
                if hasattr(widget, "is_hovered"):
                    return widget
                return None
            if hasattr(widget, "is_hovered"):
                return widget
            else:
                return first_hoverable_in_family_tree(widget.parent)
        pointed = self._get_pointed_comp()
        if pointed is None:
            self._hover(None)
        else:
            self._hover(first_hoverable_in_family_tree(pointed))


mouse = _Mouse()
del _Mouse




from baopig._debug import debug_screen_updates
from .utilities import *
from .style import Theme
from .layer import Layer
from .zone import Zone, Widget
from .selections import Selector


def decorator_open(scene, open):
    functools.wraps(open)
    def wrapped_func(*args, **kwargs):
        app = scene.application
        if app.focused_scene is scene: return

        with paint_lock:
            scene.pre_open()
            if app.focused_scene:
                app.focused_scene._close()
            app._focused_scene = scene
            app._update_display()
            scene.container_open()
            res = open(*args, **kwargs)
            scene.paint(recursive=True)

        LOGGER.debug("Open scene : {}".format(scene))
        # scene.signal.OPEN.emit()
        return res
    return wrapped_func


class Scene(Zone, Selector, Openable, Closable):
    """
    A Scene is like a page in an application. It can be the menu, the parameters page...

    If size parameter is filled, then when we swap to this scene, the application will be resized
    to 'size'. When it will swap again to another scene, if no size is required, the application
    will be resized to its original size.
    """

    STYLE = Zone.STYLE.substyle()
    STYLE.modify(
        background_color="theme-color-scene_background",
    )

    def __init__(self, application, size=None, **options):

        self.open = decorator_open(self, self.open)

        if "name" not in options:
            options["name"] = self.__class__.__name__
        self._application = self._app = application

        if "theme" in options:
            theme = options.pop("theme")
            if not isinstance(theme, str):  # theme name
                assert isinstance(theme, Theme)
                if not theme.issubtheme(application.theme):
                    raise PermissionError("Must be an application sub-theme")
        else:
            theme = application.theme.subtheme()
        self.inherit_style(theme)

        Zone.__init__(
            self,
            parent=self,
            pos=(0, 0),
            size=application.default_size if size is None else size,
            **options
        )
        Selector.__init__(self)

        # self._mode = 0
        self._asked_size = size
        self._mode_before_fullscreen = None
        self._size_before_fullscreen = None
        self._focused_comp_ref = lambda: None

        self.enable_selecting()

    def __str__(self):

        return self.name

    abs_left = 0  # End of recursive call
    abs_top = 0  # End of recursive call
    application = property(lambda self: self._application)
    asked_size = property(lambda self: self._asked_size)
    focused_comp = property(lambda self: self._focused_comp_ref())
    # mode = property(lambda self: self._mode)
    painter = property(lambda self: self._application._painter)
    scene = property(lambda self: self)  # End of recursive call

    def _add_child(self, widget):

        if widget is self:
            return self.application._add_scene(self)  # a scene is a root
        super()._add_child(widget)

    def _close(self):

        if self.application.focused_scene is not self: return
        self.container_close()
        self.close()
        Widget.set_surface(self, pygame.Surface(self.size))  # not pygame.display anymore
        self._focus(None)
        self.application._focused_scene = None

        # LOGGER.debug("Close scene : {}".format(self))

    def _dirty_child_TBR(self, child, dirty):

        if child is self:
            self._dirty = dirty
            return
        super()._dirty_child(child, dirty)

    def _focus(self, widget):

        if widget == self.focused_comp: return

        # DEFOCUS
        old_focused = self.focused_comp
        if old_focused is not None:
            assert old_focused.is_focused
            old_focused._is_focused = False

        if widget is None:
            widget = self
        else:
            assert widget.is_visible
        assert not widget.is_focused

        # FOCUS
        widget._is_focused = True
        self._focused_comp_ref = widget.get_weakref()  # (lambda: None) if widget is None else

        if old_focused is not None: old_focused.signal.DEFOCUS.emit()
        widget.signal.FOCUS.emit()
        # print("FOCUS", widget)

    def _update_rect(self):

        if self._rect_to_update is None: return
        assert self.application.focused_scene is self

        with paint_lock:

            rect = self._rect_to_update
            self._rect_to_update = None
            # for rect in rects_to_update:  # TODO : rect_to_update ? try fps
            self.surface.fill(self.background_color, rect)
            for layer in self.layers:
                for child in layer.visible:
                    if child.hitbox.colliderect(rect):
                        try:
                            collision = child.hitbox.clip(rect)
                            self.surface.blit(
                                child.surface.subsurface(
                                    (collision.left - child.rect.left, collision.top - child.rect.top) + collision.size),
                                collision.topleft
                            )
                        except pygame.error as e:
                            # can be raised from a child.surface who is a subsurface from self.surface
                            assert child.surface.get_parent() is self.surface
                            child._flip_without_update()  # overdraw child.hitbox

            if debug_screen_updates:
                LOGGER.info("update in {} :  {}".format(self, rect))

            if debug_with_assert: assert pygame.display.get_surface() is self.surface
            pygame.display.update(rect)

        if self.painter.is_recording and self.painter.is_recording.only_at_change:
            pygame.image.save(self.surface,
                              self.painter.record_directory + "record_{:0>3}.png".format(
                                  self.painter.record_index))
            self.painter.record_index += 1

    def _warn_parent(self, rect):
        raise PermissionError("Should never be called")

    def divide(self, side, width):
        raise PermissionError("Cannot divide a Scene")  # TODO : rework Zone.divide

    def pre_open(self):
        """Stuff to do right before this scene is open"""

    def receive(self, event):
        """This method is called at every pygame event"""

    def resize(self, w, h):

        if (w, h) == self.asked_size: return
        self._asked_size = (w, h)
        if self is self.application.focused_scene:
            self.application._update_display()

    def run(self):
        """Stuff to repeat endlessly while this scene is focused"""

    def set_mode_TO_REMOVE(self, mode):

        if mode is self.mode: return

        assert mode in (0, pygame.NOFRAME, pygame.RESIZABLE, pygame.FULLSCREEN)

        if mode is pygame.FULLSCREEN and self.mode != mode:
            self._mode_before_fullscreen = self.mode
            self._size_before_fullscreen = self.asked_size

            # print("Asked for fullscreen")
            # mode = 0

        self._mode = mode
        self.application._update_display()

    def toggle_debugging(self):

        if not hasattr(self, "debug_layer"):
            self.debug_layer = Layer(self, name="debug_layer", level=self.layers_manager.FOREGROUND)
            from baopig.prefabs.debugzone import DebugZone
            self.debug_zone = DebugZone(self)
        else:
            self.debug_zone.toggle_debugging()

    def toggle_fullscreen(self):

        if self.mode == pygame.FULLSCREEN:
            self.set_mode(self._mode_before_fullscreen)
        else:
            self.set_mode(pygame.FULLSCREEN)

    def asleep(self): raise PermissionError("A Scene cannot sleep")
    def wake(self): raise PermissionError("A Scene cannot sleep")
    def show(self): pass
    def hide(self): pass





import functools
from baopig.communicative import ApplicationExit, Communicative
from .widget import Widget


def decorator_start_running(widget, start_running):
    functools.wraps(start_running)
    def wrapped_func(*args, **kwargs):
        if widget.is_running is True: return
        _runables_manager.start_running(widget)
        widget._is_running = True
        res = start_running(*args, **kwargs)
        widget.signal.START_RUNNING.emit()
        return res
    return wrapped_func

def decorator_pause(widget, pause):
    functools.wraps(pause)
    def wrapped_func(*args, **kwargs):
        if widget.is_running is False: raise PermissionError("Cannot pause a Runable who didn't start yet")
        if widget.is_paused is True: return
        _runables_manager.pause(widget)
        widget._is_running = False
        widget._is_paused = True
        res = pause(*args, **kwargs)
        widget.signal.PAUSE.emit()
        return res
    return wrapped_func

def decorator_resume(widget, resume):
    functools.wraps(resume)
    def wrapped_func(*args, **kwargs):
        if widget.is_paused is False: raise PermissionError("Cannot resume a Runable who isn't paused")
        _runables_manager.resume(widget)
        widget._is_running = True
        widget._is_paused = False
        res = resume(*args, **kwargs)
        widget.signal.RESUME.emit()
        return res
    return wrapped_func

def decorator_stop_running(widget, stop_running):
    functools.wraps(stop_running)
    def wrapped_func(*args, **kwargs):
        if widget.is_paused is True: widget.resume()
        if widget.is_running is False: return
        _runables_manager.stop_running(widget)
        widget._is_running = False
        res = stop_running(*args, **kwargs)
        widget.signal.STOP_RUNNING.emit()
        return res
    return wrapped_func


class _RunablesManager:
    def __init__(self):

        self._runables = set()
        self._running = set()
        self._paused = set()

    def add(self, runable):

        assert isinstance(runable, Runable)
        self._runables.add(runable)

    def pause(self, runable):

        assert runable in self._running
        self._running.remove(runable)
        self._paused.add(runable)

    def remove(self, runable):

        assert runable in self._runables
        self._runables.remove(runable)
        if runable in self._running:
            self._running.remove(runable)

    def resume(self, runable):

        assert runable in self._paused
        self._paused.remove(runable)
        self._running.add(runable)

    def start_running(self, runable):

        assert runable in self._runables
        assert not runable in self._running
        assert not runable in self._paused
        self._running.add(runable)

    def stop_running(self, runable):

        assert runable in self._running
        self._running.remove(runable)

    def run_once(self):

        for runable in self._running:
            runable.run()


_runables_manager = _RunablesManager()
del _RunablesManager


class Runable(Communicative):
    """
    A Runable is a Communicative object with 4 signals:
        - START_RUNNING
        - PAUSE
        - RESUME
        - STOP_RUNNING
    Its 'run' method is called at each application loop
    """

    # TODO : application.runables instead of parent.children.runables, don't depend on Widget
    # -> if widget, sleeping will not run
    def __init__(self, start=False):

        _runables_manager.add(self)

        if not hasattr(self, "signal"):  # if Communicative.__init__(self) haven't been called elsewhere
            Communicative.__init__(self)

        self.start_running = decorator_start_running(self, self.start_running)
        self.pause = decorator_pause(self, self.pause)
        self.resume = decorator_resume(self, self.resume)
        self.stop_running = decorator_stop_running(self, self.stop_running)

        self._is_running = False
        self._is_paused = False

        self.create_signal("START_RUNNING")
        self.create_signal("PAUSE")
        self.create_signal("RESUME")
        self.create_signal("STOP_RUNNING")

        if isinstance(self, Widget):
        # if hasattr(self, "_dirty") and hasattr(self, "_is_sleeping"):  # if sub-classed by a Widget
            self.connect("pause", self.signal.ASLEEP)
            self.connect("resume", self.signal.WAKE)
            self.signal.KILL.connect(lambda: _runables_manager.remove(self))

        if start:
            self.start_running()

    is_paused = property(lambda self: self._is_paused)
    is_running = property(lambda self: self._is_running)

    def run(self):
        """Stuff to do when the object is running"""

    def pause(self):
        """Stuff to do when the object is paused"""

    def resume(self):
        """Stuff to do when the object resume"""

    def start_running(self):
        """Stuff to do when the object starts to run"""

    def stop_running(self):
        """Stuff to do when the object stops to run"""





from baopig import *


class UT_ColorChooserDialog(Zone):
    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)

        colorchooser = ColorChooserDialog(self.application)
        text = Text(self, "Color : None", pos=(10, 40))
        colorchooser.signal.ANSWERED.connect(lambda color: text.set_text("Color : " + str(color)))
        b = Button(self, "Wich color ?", sticky="top", command=colorchooser.open)



ut_scenes = [
    UT_ColorChooserDialog,
]

if __name__ == "__main__":
    from baopig.unit_tests.testerscene import TesterScene
    app = Application()
    for scene in ut_scenes:
        TesterScene(app, scene)
    app.launch()
